import pytest


@pytest.mark.ci
def test_setup_magnetize(separate_folder):
    from asr.setup.materials import main
    from pathlib import Path
    main()

    assert Path('materials.json').is_file()



from pathlib import Path

import numpy as np

from ase.parallel import world
from ase.io import read

from asr.core import command, option
from asr.core import write_json


def lattice_vectors(N_c):
    """Return lattice vectors for cells in the supercell."""
    # Lattice vectors relevative to the reference cell
    R_cN = np.indices(N_c).reshape(3, -1)
    N_c = np.array(N_c)[:, np.newaxis]
    R_cN += N_c // 2
    R_cN %= N_c
    R_cN -= N_c // 2

    return R_cN


@command(
    "asr.phonopy",
    requires=["structure.json", "gs.gpw"],
    dependencies=["asr.gs@calculate"],
)
@option("--n", type=int, help="Supercell size")
@option("--d", type=float, help="Displacement size")
@option('-c', '--calculator', help='Calculator params.')
def calculate(n=2, d=0.05,
              calculator={'name': 'gpaw',
                          'mode': {'name': 'pw', 'ecut': 800},
                          'xc': 'PBE',
                          'basis': 'dzp',
                          'kpts': {'density': 6.0, 'gamma': True},
                          'occupations': {'name': 'fermi-dirac',
                                          'width': 0.05},
                          'convergence': {'forces': 1.0e-4},
                          'symmetry': {'point_group': False},
                          'txt': 'phonons.txt',
                          'charge': 0}):
    """Calculate atomic forces used for phonon spectrum."""
    from asr.calculators import get_calculator

    from phonopy import Phonopy
    from phonopy.structure.atoms import PhonopyAtoms

    # Remove empty files:
    if world.rank == 0:
        for f in Path().glob("phonon.*.json"):
            if f.stat().st_size == 0:
                f.unlink()
    world.barrier()

    atoms = read("structure.json")

    from ase.calculators.calculator import get_calculator_class
    name = calculator.pop('name')
    calc = get_calculator_class(name)(**calculator)

    # Set initial magnetic moments
    from asr.core import is_magnetic

    if is_magnetic():
        gsold = get_calculator()("gs.gpw", txt=None)
        magmoms_m = gsold.get_magnetic_moments()
        atoms.set_initial_magnetic_moments(magmoms_m)

    from asr.core import get_dimensionality

    nd = get_dimensionality()
    if nd == 3:
        supercell = [[n, 0, 0], [0, n, 0], [0, 0, n]]
    elif nd == 2:
        supercell = [[n, 0, 0], [0, n, 0], [0, 0, 1]]
    elif nd == 1:
        supercell = [[n, 0, 0], [0, 1, 0], [0, 0, 1]]

    phonopy_atoms = PhonopyAtoms(symbols=atoms.symbols,
                                 cell=atoms.get_cell(),
                                 scaled_positions=atoms.get_scaled_positions())

    phonon = Phonopy(phonopy_atoms, supercell)

    phonon.generate_displacements(distance=d, is_plusminus=True)
    # displacements = phonon.get_displacements()
    displaced_sc = phonon.get_supercells_with_displacements()

    from ase.atoms import Atoms
    scell = displaced_sc[0]
    atoms_N = Atoms(symbols=scell.get_chemical_symbols(),
                    scaled_positions=scell.get_scaled_positions(),
                    cell=scell.get_cell(),
                    pbc=atoms.pbc)

    for n, cell in enumerate(displaced_sc):
        # Displacement number
        a = n // 2
        # Sign of the displacement
        sign = ["+", "-"][n % 2]

        filename = "phonons.{0}{1}.json".format(a, sign)

        if Path(filename).is_file():
            continue

        atoms_N.set_scaled_positions(cell.get_scaled_positions())
        atoms_N.set_calculator(calc)
        forces = atoms_N.get_forces()

        drift_force = forces.sum(axis=0)
        for force in forces:
            force -= drift_force / forces.shape[0]

        write_json(filename, {"force": forces})


def requires():
    return ["results-asr.phonopy@calculate.json"]


def webpanel(row, key_descriptions):
    from asr.database.browser import table, fig

    phonontable = table(row, "Property", ["minhessianeig"], key_descriptions)

    panel = {
        "title": "Phonon bandstructure",
        "columns": [[fig("phonon_bs.png")], [phonontable]],
        "plot_descriptions": [
            {"function": plot_bandstructure, "filenames": ["phonon_bs.png"]}
        ],
        "sort": 3,
    }

    dynstab = row.get("dynamic_stability_level")
    stabilities = {1: "low", 2: "medium", 3: "high"}
    high = "Min. Hessian eig. > -0.01 meV/Ang^2 AND elastic const. > 0"
    medium = "Min. Hessian eig. > -2 eV/Ang^2 AND elastic const. > 0"
    low = "Min. Hessian eig.  < -2 eV/Ang^2 OR elastic const. < 0"
    row = [
        "Phonons",
        '<a href="#" data-toggle="tooltip" data-html="true" '
        + 'title="LOW: {}&#13;MEDIUM: {}&#13;HIGH: {}">{}</a>'.format(
            low, medium, high, stabilities[dynstab].upper()
        ),
    ]

    summary = {
        "title": "Summary",
        "columns": [
            [
                {
                    "type": "table",
                    "header": ["Stability", "Category"],
                    "rows": [row],
                }
            ]
        ],
    }
    return [panel, summary]


@command(
    "asr.phonopy",
    requires=requires,
    webpanel=webpanel,
    dependencies=["asr.phonopy@calculate"],
)
def main():
    from asr.core import read_json
    from asr.core import get_dimensionality

    from phonopy import Phonopy
    from phonopy.structure.atoms import PhonopyAtoms
    from phonopy.units import THzToEv

    dct = read_json("results-asr.phonopy@calculate.json")
    atoms = read("structure.json")
    n = dct["__params__"]["n"]
    d = dct["__params__"]["d"]

    nd = get_dimensionality()
    if nd == 3:
        supercell = [[n, 0, 0], [0, n, 0], [0, 0, n]]
        N_c = (n, n, n)
    elif nd == 2:
        supercell = [[n, 0, 0], [0, n, 0], [0, 0, 1]]
        N_c = (n, n, 1)
    elif nd == 1:
        supercell = [[n, 0, 0], [0, 1, 0], [0, 0, 1]]
        N_c = (n, 1, 1)

    phonopy_atoms = PhonopyAtoms(
        symbols=atoms.symbols,
        cell=atoms.get_cell(),
        scaled_positions=atoms.get_scaled_positions(),
    )

    phonon = Phonopy(phonopy_atoms, supercell)

    phonon.generate_displacements(distance=d, is_plusminus=True)
    # displacements = phonon.get_displacements()
    displaced_sc = phonon.get_supercells_with_displacements()

    # for displace in displacements:
    #    print("[Phonopy] %d %s" % (displace[0], displace[1:]))

    set_of_forces = []

    for i, cell in enumerate(displaced_sc):
        # Displacement index
        a = i // 2
        # Sign of the diplacement
        sign = ["+", "-"][i % 2]

        filename = "phonons.{0}{1}.json".format(a, sign)

        forces = read_json(filename)["force"]
        # Number of forces equals to the number of atoms in the supercell
        assert len(forces) == len(atoms) * n ** nd, "Wrong supercell size!"

        set_of_forces.append(forces)

    phonon.produce_force_constants(
        forces=set_of_forces, calculate_full_force_constants=False
    )
    phonon.symmetrize_force_constants()

    q_qc = np.indices(N_c).reshape(3, -1).T / N_c

    omega_kl = np.zeros((len(q_qc), 3 * len(atoms)))
    u_klav = np.zeros((len(q_qc), 3 * len(atoms), len(atoms), 3))

    for q, q_c in enumerate(q_qc):
        omega_l, u_ll = phonon.get_frequencies_with_eigenvectors(q_c)
        omega_kl[q] = omega_l * THzToEv
        u_klav[q] = u_ll.reshape(3 * len(atoms), len(atoms), 3)
        if q_c.any() == 0.0:
            phonon.set_irreps(q_c)
            ob = phonon._irreps
            irreps = []
            for nr, (deg, irr) in enumerate(
                zip(ob._degenerate_sets, ob._ir_labels)
            ):
                irreps += [irr] * len(deg)

    irreps = list(irreps)

    R_cN = lattice_vectors(N_c)
    C_N = phonon.get_force_constants()
    C_N = C_N.reshape(len(atoms), len(atoms), n**nd, 3, 3)
    C_N = C_N.transpose(2, 0, 3, 1, 4)
    C_N = C_N.reshape(n**nd, 3 * len(atoms), 3 * len(atoms))

    eigs = []

    for q_c in q_qc:
        phase_N = np.exp(-2j * np.pi * np.dot(q_c, R_cN))
        C_q = np.sum(phase_N[:, np.newaxis, np.newaxis] * C_N, axis=0)
        eigs.append(np.linalg.eigvalsh(C_q))

    eigs = np.array(eigs)
    mineig = np.min(eigs)

    if mineig < -2:
        dynamic_stability = 1
    elif mineig < -1e-5:
        dynamic_stability = 2
    else:
        dynamic_stability = 3

    results = {'omega_kl': omega_kl,
               'irr_l': irreps,
               'q_qc': q_qc,
               'u_klav': u_klav,
               'minhessianeig': mineig,
               'dynamic_stability_level': dynamic_stability}

    # Next calculate an approximate phonon band structure
    nqpts = 100
    freqs_kl = np.zeros((nqpts, 3 * len(atoms)))
    path = atoms.cell.bandpath(npoints=nqpts, pbc=atoms.pbc)

    for q, q_c in enumerate(path.kpts):
        freqs = phonon.get_frequencies(q_c) * THzToEv
        freqs_kl[q] = freqs

    results['interp_freqs_kl'] = freqs_kl
    results['path'] = path
    results['__key_descriptions__'] = \
        {'minhessianeig': 'KVP: Minimum eigenvalue of Hessian [eV/Ang^2]',
         'dynamic_stability_level': 'KVP: Dynamic stability level'}

    return results


def plot_phonons(row, fname):
    import matplotlib.pyplot as plt

    data = row.data.get("results-asr.phonopy.json")
    if data is None:
        return

    omega_kl = data["omega_kl"]
    gamma = omega_kl[0]
    fig = plt.figure(figsize=(6.4, 3.9))
    ax = fig.gca()

    x0 = -0.0005  # eV
    for x, color in [(gamma[gamma < x0], "r"), (gamma[gamma >= x0], "b")]:
        if len(x) > 0:
            markerline, _, _ = ax.stem(
                x * 1000,
                np.ones_like(x),
                bottom=-1,
                markerfmt=color + "o",
                linefmt=color + "-",
            )
            plt.setp(markerline, alpha=0.4)
    ax.set_xlabel(r"phonon frequency at $\Gamma$ [meV]")
    ax.axis(ymin=0.0, ymax=1.3)
    plt.tight_layout()
    plt.savefig(fname)
    plt.close()


def plot_bandstructure(row, fname):
    from matplotlib import pyplot as plt
    from ase.dft.band_structure import BandStructure

    data = row.data.get("results-asr.phonopy.json")
    path = data["path"]
    energies = data["interp_freqs_kl"]
    bs = BandStructure(path=path, energies=energies[None, :, :], reference=0)
    bs.plot(label="Interpolated")

    exact_indices = []
    for q_c in data["q_qc"]:
        diff_kc = path.kpts - q_c
        diff_kc -= np.round(diff_kc)
        inds = np.argwhere(np.all(np.abs(diff_kc) < 1e-3, 1))
        exact_indices.extend(inds.tolist())

    en_exact = np.zeros_like(energies) + np.nan
    for ind in exact_indices:
        en_exact[ind] = energies[ind]

    bs2 = BandStructure(path=path, energies=en_exact[None])
    bs2.plot(
        ax=plt.gca(),
        ls="",
        marker="o",
        color="k",
        emin=np.min(energies * 1.1),
        emax=np.max(energies * 1.1),
        ylabel="Phonon frequencies [eV]",
        label="Exact",
    )
    plt.tight_layout()
    plt.savefig(fname)
    plt.close()


if __name__ == "__main__":
    main.cli()



from asr.core import command, option


@command('asr.dos')
@option('--name', type=str)
@option('--filename', type=str)
@option('--kptdensity', help='K point kptdensity')
def main(name='dos.gpw', filename='dos.json', kptdensity=12.0):
    """Calculate DOS."""
    from pathlib import Path
    from gpaw import GPAW
    if not Path(name).is_file():
        calc = GPAW('gs.gpw', txt='dos.txt',
                    kpts={'density': kptdensity},
                    nbands='300%',
                    convergence={'bands': -10})
        calc.get_potential_energy()
        calc.write(name)
        del calc

    calc = GPAW(name, txt=None)
    from ase.dft.dos import DOS
    dos = DOS(calc, width=0.0, window=(-5, 5), npts=1000)
    nspins = calc.get_number_of_spins()
    dosspin0_e = dos.get_dos(spin=0)
    energies_e = dos.get_energies()
    natoms = len(calc.atoms)
    volume = calc.atoms.get_volume()
    data = {'dosspin0_e': dosspin0_e.tolist(),
            'energies_e': energies_e.tolist(),
            'natoms': natoms,
            'volume': volume}
    if nspins == 2:
        dosspin1_e = dos.get_dos(spin=1)
        data['dosspin1_e'] = dosspin1_e.tolist()

    import json

    from ase.parallel import paropen
    with paropen(filename, 'w') as fd:
        json.dump(data, fd)


def collect_data(atoms):
    """Band structure PBE and GW +- SOC."""
    from ase.io.jsonio import read_json
    from pathlib import Path

    if not Path('dos.json').is_file():
        return {}, {}, {}

    dos = read_json('dos.json')

    return {}, {}, {'dos': dos}


def plot(row=None, filename='dos.png', file=None, show=False):
    """Plot DOS.

    Defaults to dos.json.
    """
    import json
    import matplotlib.pyplot as plt
    import numpy as np

    dos = None

    # Get data from row
    if row is not None:
        if 'dos' not in row.data:
            return
        dos = row.data['dos']

    # Otherwise from from file
    file = file or 'dos.json'
    if not dos:
        dos = json.load(open(file, 'r'))
    plt.figure()
    plt.plot(dos['energies_e'],
             np.array(dos['dosspin0_e']) / dos['volume'])
    plt.xlabel(r'Energy - $E_\mathrm{F}$ (eV)')
    plt.ylabel(r'DOS (states / (eV Å$^3$)')
    plt.tight_layout()
    plt.savefig(filename)
    if show:
        plt.show()
    return plt.gca()


def webpanel(row, key_descriptions):
    from asr.database.browser import fig

    panel = ('Density of states (PBE)',
             [[fig('dos.png')], []])

    things = [(plot, ['dos.png'])]

    return panel, things


if __name__ == '__main__':
    main.cli()



from asr.core import command, option, argument

import tempfile
from pathlib import Path

from ase.db import connect
from ase.db.app import app, projects
from flask import render_template, send_file
import asr
from ase import Atoms
from ase.calculators.calculator import kptdensity2monkhorstpack
from ase.geometry import cell_to_cellpar
from ase.utils import formula_metal
import warnings

tmpdir = Path(tempfile.mkdtemp(prefix="asr-app-"))  # used to cache png-files

path = Path(asr.__file__).parent.parent
app.jinja_loader.searchpath.append(str(path))


def create_key_descriptions(db=None):
    from asr.database.key_descriptions import key_descriptions
    from asr.database.fromtree import parse_kd
    from ase.db.web import create_key_descriptions

    flatten = {key: value
               for recipe, dct in key_descriptions.items()
               for key, value in dct.items()}

    if db is not None:
        metadata = db.metadata
        if 'keys' not in metadata:
            raise KeyError('Missing list of keys for database. '
                           'To fix this either: run database.fromtree again. '
                           'or python -m asr.database.set_metadata DATABASEFILE.')
        keys = metadata.get('keys')
    else:
        keys = list(flatten.keys())

    kd = {}
    for key in keys:
        description = flatten.get(key)
        if description is None:
            warnings.warn(f'Missing key description for {key}')
            continue
        kd[key] = description

    kd = {key: (desc['shortdesc'], desc['longdesc'], desc['units']) for
          key, desc in parse_kd(kd).items()}

    return create_key_descriptions(kd)


class Summary:
    def __init__(self, row, key_descriptions, create_layout,
                 subscript=None, prefix=''):
        self.row = row

        atoms = Atoms(cell=row.cell, pbc=row.pbc)
        self.size = kptdensity2monkhorstpack(atoms,
                                             kptdensity=1.8,
                                             even=False)

        self.cell = [['{:.3f}'.format(a) for a in axis] for axis in row.cell]
        par = ['{:.3f}'.format(x) for x in cell_to_cellpar(row.cell)]
        self.lengths = par[:3]
        self.angles = par[3:]

        self.stress = row.get('stress')
        if self.stress is not None:
            self.stress = ', '.join('{0:.3f}'.format(s) for s in self.stress)

        self.formula = formula_metal(row.numbers)
        if subscript:
            self.formula = subscript.sub(r'<sub>\1</sub>', self.formula)

        kd = key_descriptions
        self.layout = create_layout(row, kd, prefix)

        self.dipole = row.get('dipole')
        if self.dipole is not None:
            self.dipole = ', '.join('{0:.3f}'.format(d) for d in self.dipole)

        self.data = row.get('data')
        if self.data:
            self.data = ', '.join(self.data.keys())

        self.constraints = row.get('constraints')
        if self.constraints:
            self.constraints = ', '.join(c.__class__.__name__
                                         for c in self.constraints)


def setup_app():
    @app.route("/")
    def index():
        return render_template(
            "asr/database/templates/projects.html",
            projects=sorted(
                [
                    (name, proj["title"], proj["database"].count())
                    for name, proj in projects.items()
                ]
            ),
        )

    @app.route("/<project>/file/<uid>/<name>")
    def file(project, uid, name):
        assert project in projects
        path = tmpdir / f"{project}/{uid}-{name}"  # XXXXXXXXXXX
        return send_file(str(path))


def handle_query(args):
    return args["query"]


def row_to_dict(row, project, layout_function, tmpdir):
    from asr.database.browser import layout
    project_name = project['name']
    uid = row.get(project['uid_key'])
    s = Summary(row,
                create_layout=layout,
                key_descriptions=project['key_descriptions'],
                prefix=str(tmpdir / f'{project_name}/{uid}-'))
    return s


def initialize_project(database):
    from asr.database import browser
    from functools import partial

    db = connect(database)
    metadata = db.metadata
    name = metadata.get("name", database)

    # Make temporary directory
    (tmpdir / name).mkdir()

    metadata = db.metadata
    projects[name] = {
        "name": name,
        "title": metadata.get("title", name),
        "key_descriptions": create_key_descriptions(db),
        "uid_key": metadata.get("uid", "uid"),
        "database": db,
        "handle_query_function": handle_query,
        "row_to_dict_function": partial(
            row_to_dict, layout_function=browser.layout, tmpdir=tmpdir
        ),
        "default_columns": metadata.get("default_columns", ["formula", "uid"]),
        "search_template": str(
            metadata.get(
                "search_template", "asr/database/templates/search.html"
            )
        ),
        "row_template": str(
            metadata.get("row_template", "asr/database/templates/row.html")
        ),
    }


@command()
@argument("databases", nargs=-1)
@option("--host", help="Host address.")
@option("--test", is_flag=True, help="Test the app.")
def main(databases, host="0.0.0.0", test=False):
    for database in databases:
        initialize_project(database)

    setup_app()

    if test:
        import traceback
        app.testing = True
        with app.test_client() as c:
            for name in projects:
                print(f'Testing {name}')
                c.get(f'/{name}/').data.decode()
                project = projects[name]
                db = project['database']
                uid_key = project['uid_key']
                n = len(db)
                uids = []
                for row in db.select(include_data=False):
                    uids.append(row.get(uid_key))
                    if len(uids) == n:
                        break
                print(len(uids))

                for i, uid in enumerate(uids):
                    url = f'/{name}/row/{uid}'
                    print(f'\rRows: {i + 1}/{len(uids)} {url}',
                          end='', flush=True)
                    try:
                        c.get(url).data.decode()
                    except KeyboardInterrupt:
                        raise
                    except Exception:
                        print()
                        row = db.get(uid=uid)
                        exc = traceback.format_exc()
                        exc += (f'Problem with {uid}: '
                                f'Formula={row.formula} '
                                f'Prototype={row.crystal_prototype}\n'
                                + '-' * 20 + '\n')
                        with Path('errors.txt').open(mode='a') as fid:
                            fid.write(exc)
                            print(exc)
    else:
        app.run(host="0.0.0.0", debug=True)


if __name__ == "__main__":
    main.cli()



from collections import Counter
from typing import List, Dict, Any
from pathlib import Path

from asr.core import command, argument

from ase.db import connect
from ase.io import read
from ase.phasediagram import PhaseDiagram
from ase.db.row import AtomsRow

known_methods = ['DFT', 'DFT+D3']


def webpanel(row, key_descriptions):
    from asr.database.browser import fig, table

    hulltable1 = table(row,
                       'Stability',
                       ['hform', 'ehull'],
                       key_descriptions)
    hulltables = convex_hull_tables(row)
    panel = {'title': 'Thermodynamic stability',
             'columns': [[fig('convex-hull.png')],
                         [hulltable1] + hulltables],
             'plot_descriptions': [{'function': plot,
                                    'filenames': ['convex-hull.png']}],
             'sort': 1}

    thermostab = row.get('thermodynamic_stability_level')
    stabilities = {1: 'low', 2: 'medium', 3: 'high'}
    high = 'Heat of formation < convex hull + 0.2 eV/atom'
    medium = 'Heat of formation < 0.2 eV/atom'
    low = 'Heat of formation > 0.2 eV/atom'
    row = ['Thermodynamic',
           '<a href="#" data-toggle="tooltip" data-html="true" '
           + 'title="LOW: {}&#13;MEDIUM: {}&#13;HIGH: {}">{}</a>'.format(
               low, medium, high, stabilities[thermostab].upper())]

    summary = {'title': 'Summary',
               'columns': [[{'type': 'table',
                             'header': ['Stability', ''],
                             'rows': [row],
                             'columnwidth': 3}]],
               'sort': 1}
    return [panel, summary]


@command('asr.convex_hull',
         requires=['results-asr.structureinfo.json',
                   'results-asr.database.material_fingerprint.json'],
         dependencies=['asr.structureinfo',
                       'asr.database.material_fingerprint'],
         webpanel=webpanel)
@argument('databases', nargs=-1)
def main(databases):
    """Calculate convex hull energies.

    It is assumed that the first database supplied is the one containing the
    standard references.

    For a database to be a valid reference database each row has to have a
    "uid" key-value-pair. Additionally, it is required that the metadata of
    each database contains following keys:

        - title: Title of the reference database.
        - legend: Collective label for all references in the database to
          put on the convex hull figures.
        - name: f-string from which to derive name for a material.
        - link: f-string from which to derive an url for a material
          (see further information below).
        - label: f-string from which to derive a material specific name to
          put on convex hull figure.
        - method: String denoting the method that was used to calculate
          reference energies. Currently accepted strings: ['DFT', 'DFT+D3'].
          "DFT" means bare DFT references energies. "DFT+D3" indicate that the
          reference also include the D3 dispersion correction.

    The name and link keys are given as f-strings and can this refer to
    key-value-pairs in the given database. For example, valid metadata looks
    like:

    .. code-block:: javascript

        {
            'title': 'Bulk reference phases',
            'legend': 'Bulk',
            'name': '{row.formula}',
            'link': 'https://cmrdb.fysik.dtu.dk/oqmd12/row/{row.uid}',
            'label': '{row.formula}',
            'method': 'DFT',
        }

    Parameters
    ----------
    databases : list of str
        List of filenames of databases.

    """
    from asr.relax import main as relax
    from asr.gs import main as groundstate
    from asr.core import read_json
    atoms = read('structure.json')

    if not relax.done:
        if not groundstate.done:
            groundstate()

    # TODO: Make separate recipe for calculating vdW correction to total energy
    for filename in ['results-asr.relax.json', 'results-asr.gs.json']:
        if Path(filename).is_file():
            results = read_json(filename)
            energy = results.get('etot')
            usingd3 = results.get('__params__', {}).get('d3', False)
            break

    if usingd3:
        mymethod = 'DFT+D3'
    else:
        mymethod = 'DFT'

    formula = atoms.get_chemical_formula()
    count = Counter(atoms.get_chemical_symbols())

    dbdata = {}
    reqkeys = {'title', 'legend', 'name', 'link', 'label', 'method'}
    for database in databases:
        # Connect to databases and save relevant rows
        refdb = connect(database)
        metadata = refdb.metadata
        assert not (reqkeys - set(metadata)), \
            'Missing some essential metadata keys.'

        dbmethod = metadata['method']
        assert dbmethod in known_methods, f'Unknown method: {dbmethod}'
        assert dbmethod == mymethod, \
            ('You are using a reference database with '
             f'inconsistent methods: {mymethod} (this material) != '
             f'{dbmethod} ({database})')

        rows = []
        # Select only references which contain relevant elements
        rows.extend(select_references(refdb, set(count)))
        dbdata[database] = {'rows': rows,
                            'metadata': metadata}

    ref_energies = get_reference_energies(atoms, databases[0])
    hform = hof(energy,
                count,
                ref_energies)
    # Make a list of the relevant references
    references = []
    for data in dbdata.values():
        metadata = data['metadata']
        for row in data['rows']:
            hformref = hof(row.energy, row.count_atoms(), ref_energies)
            reference = {'hform': hformref,
                         'formula': row.formula,
                         'uid': row.uid,
                         'natoms': row.natoms}
            reference.update(metadata)
            if 'name' in reference:
                reference['name'] = reference['name'].format(row=row)
            if 'label' in reference:
                reference['label'] = reference['label'].format(row=row)
            if 'link' in reference:
                reference['link'] = reference['link'].format(row=row)
            references.append(reference)

    pdrefs = []
    for reference in references:
        h = reference['natoms'] * reference['hform']
        pdrefs.append((reference['formula'], h))

    results = {'hform': hform,
               'references': references}

    if len(count) == 1:
        ehull = hform
    else:
        pd = PhaseDiagram(pdrefs)
        e0, indices, coefs = pd.decompose(formula)
        ehull = hform - e0 / len(atoms)
        results['indices'] = indices.tolist()
        results['coefs'] = coefs.tolist()

    results['ehull'] = ehull
    results['__key_descriptions__'] = {
        'hform': 'KVP: Heat of formation [eV/atom]',
        'ehull': 'KVP: Energy above convex hull [eV/atom]',
        'thermodynamic_stability_level': 'KVP: Thermodynamic stability level'}

    if hform >= 0.2:
        thermodynamic_stability = 1
    elif hform is None or ehull is None:
        thermodynamic_stability = None
    elif ehull >= 0.2:
        thermodynamic_stability = 2
    else:
        thermodynamic_stability = 3

    results['thermodynamic_stability_level'] = thermodynamic_stability
    return results


def get_reference_energies(atoms, references):
    count = Counter(atoms.get_chemical_symbols())

    # Get reference energies
    ref_energies = {}
    refdb = connect(references)
    for row in select_references(refdb, set(count)):
        if len(row.count_atoms()) == 1:
            symbol = row.symbols[0]
            e_ref = row.energy / row.natoms
            assert symbol not in ref_energies
            ref_energies[symbol] = e_ref

    return ref_energies


def hof(energy, count, ref_energies):
    """Heat of formation."""
    energy = energy - sum(n * ref_energies[symbol]
                          for symbol, n in count.items())
    return energy / sum(count.values())


def select_references(db, symbols):
    refs: Dict[int, 'AtomsRow'] = {}

    for symbol in symbols:
        for row in db.select(symbol):
            for symb in row.count_atoms():
                if symb not in symbols:
                    break
            else:
                uid = row.get('uid')
                refs[uid] = row
    return list(refs.values())


def plot(row, fname):
    from ase.phasediagram import PhaseDiagram
    import matplotlib.pyplot as plt

    data = row.data['results-asr.convex_hull.json']

    count = row.count_atoms()
    if not (2 <= len(count) <= 3):
        return

    references = data['references']
    pdrefs = []
    legends = []
    colors = []
    for reference in references:
        h = reference['natoms'] * reference['hform']
        pdrefs.append((reference['formula'], h))
        if reference['legend'] not in legends:
            legends.append(reference['legend'])
        idlegend = legends.index(reference['legend'])
        colors.append(f'C{idlegend + 2}')

    pd = PhaseDiagram(pdrefs, verbose=False)

    fig = plt.figure(figsize=(6, 5))
    ax = fig.gca()

    if len(count) == 2:
        x, e, _, hull, simplices, xlabel, ylabel = pd.plot2d2()
        for i, j in simplices:
            ax.plot(x[[i, j]], e[[i, j]], '-', color='C0')
        names = [ref['label'] for ref in references]
        if row.hform < 0:
            mask = e < 0.05
            e = e[mask]
            x = x[mask]
            hull = hull[mask]
            names = [name for name, m in zip(names, mask) if m]
        ax.scatter(x, e, facecolor='none', marker='o', edgecolor=colors)

        delta = e.ptp() / 30
        for a, b, name, on_hull in zip(x, e, names, hull):
            va = 'center'
            ha = 'left'
            dy = 0
            dx = 0.02
            ax.text(a + dx, b + dy, name, ha=ha, va=va)

        A, B = pd.symbols
        ax.set_xlabel('{}$_{{1-x}}${}$_x$'.format(A, B))
        ax.set_ylabel(r'$\Delta H$ [eV/atom]')

        # Circle this material
        xt = count.get(B, 0) / sum(count.values())
        ax.plot([xt], [row.hform], 'o', color='C1', label='This material')
        ymin = e.min()

        ax.axis(xmin=-0.1, xmax=1.1, ymin=ymin - 2.5 * delta)
    else:
        x, y, _, hull, simplices = pd.plot2d3()
        names = [ref['label'] for ref in references]
        for i, j, k in simplices:
            ax.plot(x[[i, j, k, i]], y[[i, j, k, i]], '-', color='lightblue')
        ax.plot(x[hull], y[hull], 's', color='C0', label='On hull')
        ax.plot(x[~hull], y[~hull], 's', color='C2', label='Above hull')
        for a, b, name in zip(x, y, names):
            ax.text(a - 0.02, b, name, ha='right', va='top')
        A, B, C = pd.symbols
        plt.axis('off')

    for it, legend in enumerate(legends):
        ax.scatter([], [], facecolor='none', marker='o',
                   edgecolor=f'C{it + 2}', label=legend)

    plt.legend(loc='lower left')
    plt.tight_layout()
    plt.savefig(fname)
    plt.close()


def convex_hull_tables(row: AtomsRow) -> List[Dict[str, Any]]:
    data = row.data['results-asr.convex_hull.json']

    references = data.get('references', [])
    tables = {}
    for reference in references:
        tables[reference['title']] = []

    for reference in sorted(references, reverse=True,
                            key=lambda x: x['hform']):
        name = reference['name']
        matlink = reference['link']
        if reference['uid'] != row.uid:
            name = f'<a href="{matlink}">{name}</a>'
        e = reference['hform']
        tables[reference['title']].append([name, '{:.2f} eV/atom'.format(e)])

    final_tables = []
    for title, rows in tables.items():
        final_tables.append({'type': 'table',
                             'header': [title, ''],
                             'rows': rows})
    return final_tables


if __name__ == '__main__':
    main.cli()



import pytest
from pytest import approx


@pytest.mark.ci
def test_core(separate_folder):
    from click.testing import CliRunner
    from asr.core import command, argument, option, read_json
    from pathlib import Path
    from time import sleep

    @command("test_recipe")
    @argument("nx")
    @option("--ny", help="Optional number of y's")
    def test_recipe(nx, ny=4):
        x = [3] * nx
        y = [4] * ny
        sleep(0.1)
        return {'x': x, 'y': y}

    runner = CliRunner()
    result = runner.invoke(test_recipe._cli, ['--help'])
    assert result.exit_code == 0, result
    assert '-h, --help    Show this message and exit.' in result.output

    result = runner.invoke(test_recipe._cli, ['-h'])
    assert result.exit_code == 0
    assert '-h, --help    Show this message and exit.' in result.output

    test_recipe(nx=3)

    resultfile = Path('results-test_recipe@test_recipe.json')
    assert resultfile.is_file()

    reciperesults = read_json(resultfile)
    assert all(reciperesults["x"] == [3] * 3)
    assert all(reciperesults["y"] == [4] * 4)

    assert reciperesults["__params__"]["nx"] == 3
    assert reciperesults["__params__"]["ny"] == 4

    assert reciperesults["__resources__"]["time"] == approx(0.1, abs=0.1)
    assert reciperesults["__resources__"]["ncores"] == 1



from asr.core import command, option, argument, chdir
from asr.database.key_descriptions import key_descriptions as asr_kd


def parse_kd(key_descriptions):
    import re

    tmpkd = {}

    for key, desc in key_descriptions.items():
        descdict = {'type': None,
                    'iskvp': False,
                    'shortdesc': '',
                    'longdesc': '',
                    'units': ''}
        if isinstance(desc, dict):
            descdict.update(desc)
            tmpkd[key] = desc
            continue

        assert isinstance(desc, str), \
            f'Key description has to be dict or str. ({desc})'
        # Get key type
        desc, *keytype = desc.split('->')
        if keytype:
            descdict['type'] = keytype

        # Is this a kvp?
        iskvp = desc.startswith('KVP:')
        descdict['iskvp'] = iskvp
        desc = desc.replace('KVP:', '').strip()

        # Find units
        m = re.search(r"\[(.*)\]", desc)
        unit = m.group(1) if m else ''
        if unit:
            descdict['units'] = unit
        desc = desc.replace(f'[{unit}]', '').strip()

        # Find short description
        m = re.search(r"\!(.*)\!", desc)
        shortdesc = m.group(1) if m else ''
        if shortdesc:
            descdict['shortdesc'] = shortdesc

        # Everything remaining is the long description
        longdesc = desc.replace(f'!{shortdesc}!', '').strip()
        if longdesc:
            descdict['longdesc'] = longdesc
            if not shortdesc:
                descdict['shortdesc'] = descdict['longdesc']
        tmpkd[key] = descdict

    return tmpkd


tmpkd = parse_kd({key: value
                  for dct in asr_kd.values()
                  for key, value in dct.items()})


def get_kvp_kd(resultsdct):
    # Parse key descriptions to get long,
    # short, units and key value pairs
    key_descriptions = {}
    kvp = {}
    for key, desc in tmpkd.items():
        key_descriptions[key] = \
            (desc['shortdesc'], desc['longdesc'], desc['units'])

        if (key in resultsdct and desc['iskvp']
           and resultsdct[key] is not None):
            kvp[key] = resultsdct[key]

    return kvp, key_descriptions


def collect(filename):
    from pathlib import Path
    from asr.core import read_json
    data = {}

    # resultfile = f'results-{recipe.name}.json'
    results = read_json(filename)
    msg = f'{filename} already in data!'
    assert filename not in data, msg
    data[filename] = results

    # Find and try to collect related files for this resultsfile
    files = results.get('__files__', {})
    extra_files = results.get('__requires__', {}).copy()
    extra_files.update(results.get('__creates__', {}))

    for extrafile, checksum in extra_files.items():
        assert extrafile not in data, f'{extrafile} already collected!'

        if extrafile in files:
            continue
        file = Path(extrafile)

        if not file.is_file():
            print(f'Warning: Required file {file.absolute()}'
                  ' doesn\'t exist.')
            continue

        if file.suffix == '.json':
            dct = read_json(extrafile)
        else:
            dct = {'pointer': str(file.absolute())}
        data[extrafile] = dct

    links = results.get('__links__', {})
    kvp, key_descriptions = get_kvp_kd(results)
    return kvp, key_descriptions, data, links


tests = [
    {'cli': ['asr run setup.materials',
             ('asr run "database.totree materials.json --run'
              ' --atomsname structure.json"'),
             'asr run "database.fromtree tree/*/*/*/"',
             ('asr run "database.totree database.db '
              '-t newtree/{formula} --run"')],
     'results': [{'file': 'newtree/Ag/structure.json'}]},
    {'cli': ['asr run setup.materials',
             'asr run "database.totree materials.json -s natoms<2 --run'
             ' -t tree/{formula} --atomsname structure.json"',
             'asr run structureinfo tree/*/',
             'asr run "database.fromtree tree/*/"',
             'asr run "database.totree database.db '
             '-t newtree/{formula} --run"'],
     'results': [{'file': 'newtree/Ag/structure.json'},
                 {'file': 'newtree/Ag/results-asr.structureinfo.json'}]}
]


@command('asr.database.fromtree',
         tests=tests)
@argument('folders', nargs=-1)
@option('--patterns', help='Only select files matching pattern.')
@option('--dbname', help='Database name.')
@option('-m', '--metadata-from-file', help='Get metadata from file.')
def main(folders=None, patterns='info.json,results-asr.*.json',
         dbname='database.db', metadata_from_file=None):
    """Collect ASR data from folder tree into an ASE database."""
    from ase.db import connect
    from ase.io import read
    from asr.core import read_json
    import glob
    from pathlib import Path
    from fnmatch import fnmatch
    from asr.database.material_fingerprint import main as mf
    from gpaw.mpi import world

    def item_show_func(item):
        return str(item)

    atomsname = 'structure.json'
    if not folders:
        folders = ['.']
    else:
        tmpfolders = []
        for folder in folders:
            tmpfolders.extend(glob.glob(folder))
        folders = tmpfolders

    folders.sort()
    patterns = patterns.split(',')
    # We use absolute path because of chdir below!
    dbpath = Path(dbname).absolute()
    metadata = {}
    if metadata_from_file:
        metadata.update(read_json(metadata_from_file))

    if world.size > 1:
        mydbname = dbpath.parent / f'{dbname}.{world.rank}.db'
        myfolders = folders[world.rank::world.size]
    else:
        mydbname = str(dbpath)
        myfolders = folders

    nfolders = len(myfolders)
    keys = set()
    with connect(mydbname, serial=True) as db:
        for ifol, folder in enumerate(myfolders):
            if world.size > 1:
                print(f'Collecting folder {folder} on rank {world.rank} '
                      f'({ifol + 1}/{nfolders})',
                      flush=True)
            else:
                print(f'Collecting folder {folder} ({ifol}/{nfolders})',
                      flush=True)
            with chdir(folder):
                kvp = {'folder': str(folder)}
                data = {'__links__': {}}

                if not Path(atomsname).is_file():
                    continue

                if not mf.done:
                    mf()

                atoms = read(atomsname, parallel=False)
                data[atomsname] = read_json(atomsname)
                for filename in glob.glob('*'):
                    for pattern in patterns:
                        if fnmatch(filename, pattern):
                            break
                    else:
                        continue
                    tmpkvp, tmpkd, tmpdata, tmplinks = \
                        collect(str(filename))
                    if tmpkvp or tmpkd or tmpdata or tmplinks:
                        kvp.update(tmpkvp)
                        data.update(tmpdata)
                        data['__links__'].update(tmplinks)

            for key in filter(lambda x: x.startswith('results-'), data.keys()):
                recipe = key[8:-5].replace('.', '_').replace('@', '_')
                name = f'has_{recipe}'
                kvp[name] = True

            keys.update(kvp.keys())
            db.write(atoms, data=data, **kvp)

    metadata['keys'] = sorted(list(keys))
    db.metadata = metadata

    if world.size > 1:
        # Then we have to collect the separately collected databases
        # to a single final database file.
        world.barrier()
        if world.rank == 0:
            print(f'Merging separate database files to {dbname}',
                  flush=True)
            nmat = 0
            keys = set()
            with connect(dbname, serial=True) as db2:
                for rank in range(world.size):
                    dbrankname = f'{dbname}.{rank}.db'
                    print(f'Merging {dbrankname} into {dbname}', flush=True)
                    with connect(f'{dbrankname}', serial=True) as db:
                        for row in db.select():
                            kvp = row.get('key_value_pairs', {})
                            db2.write(row, data=row.get('data'), **kvp)
                            nmat += 1
                    keys.update(set(db.metadata['keys']))

            print('Done. Setting metadata.', flush=True)
            metadata['keys'] = sorted(list(keys))
            db2.metadata = metadata
            nmatdb = len(db2)
            assert nmatdb == nmat, \
                ('Merging of databases went wrong, '
                 f'number of materials changed: {nmatdb} != {nmat}')


if __name__ == '__main__':
    main.cli()



def eigenvalues(calc):
    """Get eigenvalues from calculator.

    Parameters
    ----------
    calc : Calculator

    Returns
    -------
    e_skn: (ns, nk, nb)-shape array
    """
    import numpy as np
    rs = range(calc.get_number_of_spins())
    rk = range(len(calc.get_ibz_k_points()))
    e = calc.get_eigenvalues
    return np.asarray([[e(spin=s, kpt=k) for k in rk] for s in rs])


def fermi_level(calc, eps_skn=None, nelectrons=None):
    """Get Fermi level from calculation.

    Parameters
    ----------
        calc : GPAW
            GPAW calculator
        eps_skn : ndarray, shape=(ns, nk, nb), optional
            eigenvalues (taken from calc if None)
        nelectrons : float, optional
            number of electrons (taken from calc if None)

    Returns
    -------
        out : float
            fermi level
    """
    from gpaw.occupations import occupation_numbers
    from ase.units import Ha
    if nelectrons is None:
        nelectrons = calc.get_number_of_electrons()
    if eps_skn is None:
        eps_skn = eigenvalues(calc)
    eps_skn.sort(axis=-1)
    occ = calc.occupations.todict()
    weight_k = calc.get_k_point_weights()
    return occupation_numbers(occ, eps_skn, weight_k, nelectrons)[1] * Ha


def calc2eigs(calc, ranks, soc=True, bands=None, return_spin=False,
              theta=0, phi=0):
    from gpaw.spinorbit import get_spinorbit_eigenvalues
    from gpaw import mpi
    from ase.parallel import broadcast
    import numpy as np
    from .symmetry import restrict_spin_projection_2d
    from gpaw.symmetry import atoms2symmetry

    dct = None
    if mpi.world.rank in ranks:
        if bands is None:
            n2 = calc.todict().get("convergence", {}).get("bands")
            if isinstance(n2, str):
                assert n2.startswith('CBM')
                n2 = calc.get_number_of_bands()
            bands = slice(0, n2)
        if isinstance(bands, slice):
            bands = range(calc.get_number_of_bands())[bands]
        eps_nosoc_skn = eigenvalues(calc)[..., bands]
        efermi_nosoc = calc.get_fermi_level()
        dct = {'eps_nosoc_skn': eps_nosoc_skn,
               'efermi_nosoc': efermi_nosoc}
        if soc:
            eps_mk, s_kvm = get_spinorbit_eigenvalues(calc, bands=bands,
                                                      theta=theta,
                                                      phi=phi,
                                                      return_spin=True)
            eps_km = eps_mk.T
            efermi = fermi_level(calc, eps_km[np.newaxis],
                                 nelectrons=2 * calc.get_number_of_electrons())
            symmetry = atoms2symmetry(calc.atoms)
            ibzk_kc = calc.get_ibz_k_points()
            # For magnetic systems we know some more about the spin
            # projections
            if calc.get_number_of_spins() == 1:
                # Inversion + time reversal symmetry forces degenerates spins
                if symmetry.has_inversion:
                    s_kvm[:] = 0.0
                else:
                    # For 2D we try to find materials where
                    # spins are restricted to inplane spins
                    if np.sum(calc.atoms.pbc).astype(int) == 2:
                        for ik, kpt in enumerate(ibzk_kc):
                            s_vm = restrict_spin_projection_2d(kpt,
                                                               symmetry.op_scc,
                                                               s_kvm[ik])
                            s_kvm[ik] = s_vm
            dct.update({'eps_km': eps_km,
                        'efermi': efermi,
                        's_kvm': s_kvm})

    dct = broadcast(dct, root=0, comm=mpi.world)
    if soc is None:
        raise NotImplementedError('soc=None is not implemented')

    if soc:
        out = (dct['eps_km'], dct['efermi'], dct['s_kvm'])
        if not return_spin:
            out = out[:2]
        return out
    else:
        if not return_spin:
            return dct['eps_nosoc_skn'], dct['efermi_nosoc']
        return dct['eps_nosoc_skn'], dct['efermi_nosoc'], dct['s_kvm']


def gpw2eigs(gpw, soc=True, bands=None, return_spin=False,
             theta=0, phi=0):
    """Give the eigenvalues w or w/o spinorbit coupling and the corresponding fermi energy.

    Parameters
    ----------
        gpw (str): gpw filename
        soc : None, bool
            use spinorbit coupling if None it returns both w and w/o
        bands : slice, list of ints or None
            None gives parameters.convergence.bands if possible else all bands

    Returns
    -------
    dict or e_skn, efermi
        containg eigenvalues and fermi levels w and w/o spinorbit coupling
    """
    from gpaw import GPAW
    from gpaw import mpi
    ranks = [0]
    calc = GPAW(gpw, txt=None, communicator=mpi.serial_comm)
    return calc2eigs(calc, soc=soc, bands=bands, return_spin=return_spin,
                     theta=theta, phi=phi,
                     ranks=ranks)



from asr.core import command, option


class MaterialNotFoundError(Exception):
    pass


class DBAlreadyExistsError(Exception):
    pass


class ParseError(Exception):
    pass


def where(pred, ls):
    return list(filter(pred, ls))


def only(pred, ls):
    rs = where(pred, ls)
    assert len(rs) == 1
    return rs[0]


def single(pred, ls):
    rs = where(pred, ls)
    N = len(rs)
    assert N == 0 or N == 1
    if N == 1:
        return rs[0]
    else:
        return None


def select(pred, ls):
    return list(map(pred, ls))


def count(pred, ls):
    rs = select(pred, ls)
    return sum(rs)


def unique(ls, selector=None):
    if selector:
        rs = select(selector, ls)
    else:
        rs = ls

    return all(count(lambda x: x == y, rs) == 1 for y in rs)


def parse_reactions(reactionsstr):
    import re

    with open(reactionsstr, "r") as f:
        data = f.read()

    lines = [line for line in data.split("\n") if line != ""]
    reactions = []

    splitter_re = r"(([A-Z]+[a-z]*[0-9]*)+)(\s)+([-+]?[0-9]+(\.[0-9]*)?)"
    for line in lines:
        tline = line.strip()
        match = re.match(splitter_re, tline)
        if match:
            form = match.group(1)
            energy = float(match.group(4))
            reactions.append((form, energy))
        else:
            raise ParseError(
                "Could not parse line" + ' "{}" in {}'.format(line, reactionsstr)
            )

    if not unique(reactions, lambda t: t[0]):
        bad = where(
            lambda y: count(lambda x: x[0] == y[1][0], reactions) > 1,
            enumerate(reactions),
        )
        raise ParseError("Same reaction was entered" + "multiple times: {}".format(bad))
    return reactions


def parse_refs(refsstr):
    import re

    with open(refsstr, "r") as f:
        data = f.read()

    lines = [line for line in data.split("\n") if line != ""]
    refs = []

    parser_re = r"(^[A-Z]+[a-z]*[0-9]*$)"
    for line in lines:
        tline = line.strip()
        match = re.match(parser_re, tline)
        if match:
            form = match.group(1)
            refs.append(form)
        else:
            raise ParseError(
                "Could not parse line" + ' "{}" in {}'.format(line, refsstr)
            )

    if not unique(refs):
        bad = where(lambda y: count(lambda x: x == y[1], refs) > 1, enumerate(refs))
        raise ParseError(
            "Same reference" + "was entered multiple times: {}".format(bad)
        )
    return refs


def load_data(reactionsstr, refsstr):
    reacts = parse_reactions(reactionsstr)
    refs = parse_refs(refsstr)
    return reacts, refs


def elements_from_refs(refs):
    from ase.formula import Formula

    els = []
    for ref in refs:
        el = only(lambda t: True, Formula(ref).count().keys())
        els.append(el)
    return els


def multiply_formula(prod, j):
    from ase.formula import Formula

    form = Formula(prod)
    return Formula.from_dict({k: v * j for k, v in form.count().items()})


def safe_get(db, prod):
    result = None
    for j in range(20):
        formula = multiply_formula(prod, j + 1)
        try:
            result = db.get("formula={}".format(formula))
            break
        except KeyError:
            continue

    if result is None:
        raise MaterialNotFoundError("Could not find {} in db".format(prod))

    return result


def get_hof(db, formula):
    from ase.formula import Formula

    elements = list(formula.count().keys())
    row = safe_get(db, str(formula))
    dbformula = Formula(str(row.formula))
    hof = row.energy
    for el in elements:
        elrow = safe_get(db, el)
        elformula = Formula(elrow.formula)
        number_el = only(lambda t: True, elformula.count().values())
        factor = dbformula.count()[el] / number_el
        hof -= factor * elrow.energy

    num_atoms = sum(dbformula.count().values())
    return hof / num_atoms


def get_dE_alpha(db, reactions, refs):
    from ase.formula import Formula
    from scipy import sparse

    alpha = sparse.lil_matrix((len(reactions), len(refs)))
    DE = sparse.lil_matrix((len(reactions), 1))

    for i1, (form, eexp) in enumerate(reactions):
        formula = Formula(form)
        hof = get_hof(db, formula)

        DE[i1, 0] = eexp - hof

        num_atoms = sum(formula.count().values())
        for i2, ref in enumerate(refs):
            reff = Formula(ref)
            el = only(lambda t: True, reff.count().keys())
            if el in formula.count().keys():
                alpha[i1, i2] = formula.count()[el] / num_atoms

    return DE, alpha


def minimize_error(dE, alpha):
    from scipy.sparse.linalg import spsolve
    import numpy as np

    b = -alpha.T.dot(dE)
    A = alpha.T.dot(alpha)

    dMu = spsolve(A, b)

    d = alpha.dot(dMu)
    error = dE.T.dot(dE) + 2 * dE.T.dot(alpha.dot(dMu)) + d.T.dot(d)
    error = np.sqrt(error / dE.shape[0])

    return dMu, error


def formulas_eq(form1, form2):
    if type(form1) == str:
        from ase.formula import Formula

        form1 = Formula(form1)
    if type(form2) == str:
        from ase.formula import Formula

        form2 = Formula(form2)
    return form1.stoichiometry()[:-1] == form2.stoichiometry()[:-1]


def create_corrected_db(newname, db, reactions, els_dMu):
    from ase.formula import Formula
    from ase.db import connect

    newdb = connect(newname)

    for row in db.select():
        formula = Formula(row.formula)
        el_dmu = single(lambda t: formulas_eq(t[0], formula), els_dMu)
        if el_dmu:
            el, dmu = el_dmu
            row.energy += formula.count()[el] * dmu
        newdb.write(row)


@command("asr.fere", resources="1:1h")
@option("--newdbname", help="Name of the new db file")
@option("--dbname", help="Name of the base db file")
@option(
    "--reactionsname", help="File containing reactions and energies with which to fit"
)
@option(
    "--referencesname",
    help="File containing the elements"
    + " whose references energies should be adjusted",
)
def main(
    newdbname="newdb.db",
    dbname="db.db",
    reactionsname="reactions.txt",
    referencesname="references.txt",
):
    from ase.db import connect
    import os
    import numpy as np

    if os.path.exists(newdbname):
        raise DBAlreadyExistsError
    reactions, refs = load_data(reactionsname, referencesname)

    db = connect(dbname)
    dE, alpha = get_dE_alpha(db, reactions, refs)

    dMu, error = minimize_error(dE, alpha)

    elements = elements_from_refs(refs)
    create_corrected_db(newdbname, db, reactions, list(zip(elements, dMu)))

    results = {
        "dbname": dbname,
        "newdbname": newdbname,
        "reactions": reactions,
        "refs": refs,
        "dE": np.array(dE.todense()),
        "alpha": str(alpha),
        "dMu": dMu,
        "error": error,
    }

    results["__key_descriptions__"] = {
        "dbname": "Name of base db",
        "newdbname": "Name of corrected db",
        "reactions": "Reactions and energies used to correct",
        "refs": "References that were adjusted",
        "dE": "Difference between target and initial HoFs",
        "alpha": "Alpha matrix",
        "dMu": "Adjustment of reference energies",
        "error": "RMSE after adjustment",
    }

    return results


if __name__ == "__main__":
    main.cli()



from types import SimpleNamespace


def broadcast(n, root):
    pass


def new_communicator(ranks):
    pass


def barrier():
    pass


world = SimpleNamespace(size=1,
                        rank=0,
                        broadcast=broadcast,
                        barrier=barrier,
                        new_communicator=new_communicator)

serial_comm = None

SerialCommunicator = SimpleNamespace



from .conftest import test_materials, get_webcontent
from ase.units import Bohr
import numpy as np
import pytest


def zero_pad_non_pbc_strain_directions(matrix_cvv, pbc_c):
    return matrix_cvv * pbc_c[None] * pbc_c[None, None]


def get_strain_from_atoms(inv_cell_vc, atoms):
    return np.dot(inv_cell_vc, atoms.get_cell() / Bohr) - np.eye(3)


@pytest.mark.ci
@pytest.mark.parametrize("atoms", test_materials)
@pytest.mark.parametrize("nspins", [1, 2])
def test_piezoelectrictensor(separate_folder, mockgpaw, mocker, atoms, nspins):
    import numpy as np
    from gpaw import GPAW

    cell_cv = atoms.get_cell() / Bohr
    pbc_c = atoms.get_pbc()
    inv_cell_vc = np.linalg.inv(cell_cv)
    # dphase_c / dstrain_vv
    dpde_cvv = zero_pad_non_pbc_strain_directions(
        np.array([np.eye(3)] * 3),
        pbc_c,
    )

    # ddipol_v / dstrain_vv
    ddipolde_vvv = zero_pad_non_pbc_strain_directions(
        np.array([np.eye(3)] * 3),
        pbc_c,
    )

    # Also move atomic positions
    natoms = len(atoms)
    dsposde_acvv = np.zeros((natoms, 3, 3, 3), float)
    dsposde_acvv[:, :] = np.eye(3)
    dsposde_acvv *= pbc_c[None, None] * pbc_c[None, None, None]
    spos_ac = atoms.get_scaled_positions(wrap=True)

    def _get_berry_phases(self, dir=0, spin=0):
        strain_vv = get_strain_from_atoms(inv_cell_vc, self.atoms)
        phase_c = np.dot(dpde_cvv.reshape(3, -1), strain_vv.reshape(-1))
        return [phase_c[dir]]

    def _get_scaled_positions(self):
        strain_vv = get_strain_from_atoms(inv_cell_vc, self.atoms)
        dspos_ac = np.dot(dsposde_acvv.reshape(natoms, 3, -1),
                          strain_vv.reshape(-1))
        return spos_ac + dspos_ac

    def _get_dipole_moment(self):
        strain_vv = get_strain_from_atoms(inv_cell_vc, self.atoms)
        dipol_v = np.dot(ddipolde_vvv.reshape(3, -1), strain_vv.reshape(-1))
        return dipol_v

    def _get_setup_nvalence(self, element_number):
        return 1

    def get_number_of_spins(self):
        return nspins

    mocker.patch.object(GPAW, '_get_berry_phases', new=_get_berry_phases)
    mocker.patch.object(GPAW, '_get_dipole_moment', new=_get_dipole_moment)
    mocker.patch.object(GPAW, '_get_scaled_positions', new=_get_scaled_positions)
    mocker.patch.object(GPAW, '_get_setup_nvalence', new=_get_setup_nvalence)
    mocker.patch.object(GPAW, 'get_number_of_spins', new=get_number_of_spins)
    from ase.io import write
    from asr.piezoelectrictensor import main
    write('structure.json', atoms)
    results = main()
    content = get_webcontent('database.db')

    N = np.abs(np.linalg.det(cell_cv[~pbc_c][:, ~pbc_c]))
    vol = atoms.get_volume() / Bohr**3

    # Formula for piezoeletric tensor
    # (The last factor of 2 is for spins)
    eps_berry_analytic_vvv = np.tensordot(
        cell_cv, dpde_cvv,
        axes=([0], [0])
    ) / (2 * np.pi * vol) * N * 2

    eps_dipole_analytic_vvv = ddipolde_vvv / (vol * Bohr) * N

    Z_a = [1] * natoms
    eps_atomic_analytic_vvv = np.tensordot(
        cell_cv,
        np.tensordot(
            Z_a,
            dsposde_acvv,
            axes=([0], [0]),
        ),
        axes=([0], [0])
    ) / vol * N

    eps_analytic_vvv = np.zeros((3, 3, 3), float)
    eps_analytic_vvv[pbc_c] = (eps_berry_analytic_vvv[pbc_c]
                               + eps_atomic_analytic_vvv[pbc_c])
    eps_analytic_vvv[~pbc_c] = eps_dipole_analytic_vvv[~pbc_c]

    eps_vvv = results['eps_vvv']
    eps_clamped_vvv = results['eps_clamped_vvv']
    assert eps_vvv == pytest.approx(eps_analytic_vvv)
    assert eps_clamped_vvv == pytest.approx(eps_analytic_vvv)
    assert "Piezoelectrictensor" in content, content


@pytest.mark.acceptance_test
def test_piezo_BN(separate_folder):
    from .conftest import BN
    from asr.piezoelectrictensor import main
    from asr.setup.params import main as setupparams

    formal_calculator = {
        'name': 'gpaw',
        'mode': {'name': 'pw', 'ecut': 300},
        'xc': 'PBE',
        'basis': 'dzp',
        'kpts': {'density': 2.0},
        'occupations': {'name': 'fermi-dirac',
                        'width': 0.05},
        'txt': 'formalpol.txt',
        'charge': 0
    }
    setupparams(params={'asr.formalpolarization': {'calculator': formal_calculator}})
    BN.write('structure.json')
    results = main()

    full_response = 0.12  # From a default parameter calculation of BN
    clamped_response = 0.044  # From a default parameter calculation of BN
    for eps, response in [('eps_vvv', full_response),
                          ('eps_clamped_vvv', clamped_response)]:
        eps_calculated_vvv = results[eps]
        eps_converged_vvv = np.array([[[0.0, response, 0.0],
                                       [response, 0.0, 0.0],
                                       [0.0, 0.0, 0.0]],
                                      [[response, 0.0, 0.0],
                                       [0.0, -response, 0.0],
                                       [0.0, 0.0, 0.0]],
                                      [[0.0, 0.0, 0.0],
                                       [0.0, 0.0, 0.0],
                                       [0.0, 0.0, 0.0]]])

        assert eps_calculated_vvv == pytest.approx(eps_converged_vvv, rel=0.05)



"""Module for generating atomic structures with displaced atoms.

The main recipe of this module is :func:`asr.setup.displacements.main`

.. autofunction:: asr.setup.displacements.main
"""

from pathlib import Path
from asr.core import command, option


def get_displacement_folder(atomic_index,
                            cartesian_index,
                            displacement_sign,
                            displacement):
    """Generate folder name from (ia, iv, sign, displacement)."""
    cartesian_symbol = 'xyz'[cartesian_index]
    displacement_symbol = ' +-'[displacement_sign]
    foldername = (f'{displacement}-{atomic_index}'
                  f'-{displacement_symbol}{cartesian_symbol}')
    folder = Path('displacements') / foldername
    return folder


def create_displacements_folder(folder):
    folder.mkdir(parents=True, exist_ok=False)


def get_all_displacements(atoms):
    """Generate ia, iv, sign for all displacements."""
    for ia in range(len(atoms)):
        for iv in range(3):
            for sign in [-1, 1]:
                yield (ia, iv, sign)


def displace_atom(atoms, ia, iv, sign, delta):
    new_atoms = atoms.copy()
    pos_av = new_atoms.get_positions()
    pos_av[ia, iv] += sign * delta
    new_atoms.set_positions(pos_av)
    return new_atoms


@command('asr.setup.displacements')
@option('--displacement', help='How much to displace atoms.')
@option('--copy-params', help='Copy params.json to displacement folders.')
def main(displacement=0.01, copy_params=True):
    """Generate atomic displacements.

    Generate atomic structures with displaced atoms. The generated
    atomic structures are written to 'structure.json' and put into a
    directory with the structure

        displacements/{displacement}-{atomic_index}-{displacement_symbol}{cartesian_symbol}

    Notice that all generated directories are a sub-directory of displacements/.

    """
    from ase.io import read
    from ase.parallel import world
    structure = read('structure.json')
    folders = []
    params = Path('params.json')
    if not params.is_file():
        copy_params = False
    else:
        params_text = params.read_text()

    for ia, iv, sign in get_all_displacements(structure):
        folder = get_displacement_folder(ia, iv,
                                         sign, displacement)
        if world.rank == 0:
            create_displacements_folder(folder)
        new_structure = displace_atom(structure, ia, iv, sign, displacement)
        new_structure.write(folder / 'structure.json')
        folders.append(str(folder))

        if copy_params and params.is_file() and world.rank == 0:
            (folder / 'params.json').write_text(params_text)

    world.barrier()
    return {'folders': folders}



from click import Choice
from asr.core import command, option


@command('asr.setup.magnetize')
@option('--state', type=Choice(['all', 'nm', 'fm', 'afm']),
        help='Magnetic states to create.')
@option('--name', help='Atomic structure')
@option('--copy-params', is_flag=True,
        help='Also copy params.json from this dir (if exists).')
def main(state='all', name='unrelaxed.json', copy_params=False):
    """Set up folders with magnetic moments.

    This recipe can be used to test different magnetic configurations
    of an atomic structure. The recipe creates new folder using acronyms
    for the magnetic configutions. Supported magnetic configurations at the
    moment:

    * nm/  <- non-magnetic
    * fm/  <- ferro-magnetic
    * afm/ <- anti-ferro-magnetic (only works with exactly two-magnetic atoms
      in the unit cell)

    To test a specific magnetic configution, for example ferromagnetic,
    simply use the --state switch: --state fm.

    Use the 'all' identifier to test all magnetic configuration: --state all
    This is the default option.

    An atomic structure with the correct initial magnetic moments
    named 'unrelaxed.json' is saved in each folder, which can be relaxed
    using the relax recipe.

    By default this recipe looks for a structure in the current directory named
    'unrelaxed.json'. Use the --name switch to change this name.

    If you also want to copy the params.json file in the current directory into
    all newly created directories use the --copy-params switch.

    Examples
    --------
    Set up all known magnetic configurations (assuming existence of 'unrelaxed.json')
    >>> asr run setup.magnetize

    Only set up ferromagnetic configuration
    >>> asr run "setup.magnetic --state fm"

    """
    from pathlib import Path
    from ase.io import read, write
    from ase.parallel import world
    from asr.core import magnetic_atoms
    import numpy as np
    known_states = ['nm', 'fm', 'afm']

    if state == 'all':
        states = known_states
    else:
        states = [state]

    for state in states:
        msg = f'{state} is not a known state!'
        assert state in known_states, msg

    # Non-magnetic:
    if 'nm' in states:
        atoms = read(name)
        atoms.set_initial_magnetic_moments(None)
        assert not Path('nm').is_dir(), 'nm/ already exists!'
        if world.rank == 0:
            Path('nm').mkdir()
            write('nm/unrelaxed.json', atoms)
            if copy_params:
                p = Path('params.json')
                if p.is_file:
                    Path('nm/params.json').write_text(p.read_text())

    # Ferro-magnetic:
    if 'fm' in states:
        atoms = read(name)
        atoms.set_initial_magnetic_moments([1] * len(atoms))
        assert not Path('fm').is_dir(), 'fm/ already exists!'
        if world.rank == 0:
            Path('fm').mkdir()
            write('fm/unrelaxed.json', atoms)
            if copy_params:
                p = Path('params.json')
                if p.is_file:
                    Path('fm/params.json').write_text(p.read_text())

    # Antiferro-magnetic:
    if 'afm' in states:
        atoms = read(name)
        magnetic = magnetic_atoms(atoms)
        nmag = sum(magnetic)
        if nmag == 2:
            magmoms = np.zeros(len(atoms))
            a1, a2 = np.where(magnetic)[0]
            magmoms[a1] = 1.0
            magmoms[a2] = -1.0
            atoms.set_initial_magnetic_moments(magmoms)
            assert not Path('afm').is_dir(), 'afm/ already exists!'
            if world.rank == 0:
                Path('afm').mkdir()
                write('afm/unrelaxed.json', atoms)
                if copy_params:
                    p = Path('params.json')
                    if p.is_file:
                        Path('afm/params.json').write_text(p.read_text())
        else:
            print('Warning: Did not produce afm state. '
                  f'The number of magnetic atoms is {nmag}. '
                  'At the moment, I only know how to do AFM '
                  'state for 2 magnetic atoms.')


if __name__ == '__main__':
    main.cli()



from asr.core import command, argument, get_recipes


# Style: "KVP: Long description !short description! [unit]

key_descriptions = {
    "berry": {"Topology": "KVP: Band topology !Topology!"},
    "bse": {"E_B": "KVP: Exciton binding energy from BSE "
            "!Exc. bind. energy! [eV]"},
    "convex_hull": {
        "ehull": "KVP: Energy above convex hull [eV/atom]",
        "hform": "KVP: Heat of formation [eV/atom]",
        "thermodynamic_stability_level": "KVP: Thermodynamic stability level",
    },
    "gs": {
        "forces": "Forces on atoms [eV/Angstrom]",
        "stresses": "Stress on unit cell [`eV/Angstrom^{dim-1}`]",
        "etot": "KVP: Total energy !Tot. En.! [eV]",
        "evac": "KVP: Vacuum level !Vacuum level! [eV]",
        "evacdiff":
        "KVP: Vacuum level difference !Vacuum level difference! [eV]",
        "dipz": "KVP: Out-of-plane dipole along +z axis [e * Ang]",
        "efermi": "KVP: Fermi level !Fermi level! [eV]",
        "gap": "KVP: Band gap !Band gap! [eV]",
        "vbm": "KVP: Valence band maximum !Val. band maximum! [eV]",
        "cbm": "KVP: Conduction band minimum !Cond. band maximum! [eV]",
        "gap_dir": "KVP: Direct band gap !Dir. band gap! [eV]",
        "gap_dir_nosoc":
        "KVP: Direct gap w/o soc. !Dir. gap wo. soc.! [eV]",
        "workfunction": "KVP: Work function (avg. if finite dipole) [eV]",
    },
    "gw": {
        "vbm_gw_nosoc": "Valence band maximum w/o soc. (GW) [eV]",
        "cbm_gw_nosoc": "Conduction band minimum w/o soc. (GW) [eV]",
        "gap_dir_gw_nosoc": "Direct gap w/o soc. (GW) [eV]",
        "gap_gw_nosoc": "Gap w/o soc. (GW) [eV]",
        "kvbm_nosoc": "k-point of GW valence band maximum w/o soc",
        "kcbm_nosoc": "k-point of GW conduction band minimum w/o soc",
        "vbm_gw": "KVP: Valence band maximum (GW) [eV]",
        "cbm_gw": "KVP: Conduction band minimum (GW) [eV]",
        "gap_dir_gw": "KVP: Direct band gap (GW) [eV]",
        "gap_gw": "KVP: Band gap (GW) [eV]",
        "kvbm": "k-point of GW valence band maximum",
        "kcbm": "k-point of GW conduction band minimum",
        "efermi_gw_nosoc": "Fermi energy w/o soc. (GW) [eV]",
        "efermi_gw_soc": "Fermi energy (GW) [eV]",
    },
    "hse": {
        "vbm_hse_nosoc": "Valence band maximum w/o soc. (HSE) [eV]",
        "cbm_hse_nosoc": "Conduction band minimum w/o soc. (HSE) [eV]",
        "gap_dir_hse_nosoc": "Direct gap w/o soc. (HSE) [eV]",
        "gap_hse_nosoc": "Band gap w/o soc. (HSE) [eV]",
        "kvbm_nosoc": "k-point of HSE valence band maximum w/o soc",
        "kcbm_nosoc": "k-point of HSE conduction band minimum w/o soc",
        "vbm_hse": "KVP: Valence band maximum (HSE) [eV]",
        "cbm_hse": "KVP: Conduction band minimum (HSE) [eV]",
        "gap_dir_hse": "KVP: Direct band gap (HSE) [eV]",
        "gap_hse": "KVP: Band gap (HSE) [eV]",
        "kvbm": "k-point of HSE valence band maximum",
        "kcbm": "k-point of HSE conduction band minimum",
        "efermi_hse_nosoc": "Fermi energy w/o soc. (HSE) [eV]",
        "efermi_hse_soc": "Fermi energy (HSE) [eV]",
    },
    "infraredpolarizability": {
        "alphax_lat": "KVP: Static lattice polarizability (x) [Ang]",
        "alphay_lat": "KVP: Static lattice polarizability (y) [Ang]",
        "alphaz_lat": "KVP: Static lattice polarizability (z) [Ang]",
        "alphax": "KVP: Static total polarizability (x) [Ang]",
        "alphay": "KVP: Static total polarizability (y) [Ang]",
        "alphaz": "KVP: Static total polarizability (z) [Ang]",
    },
    "magnetic_anisotropy": {
        "spin_axis": "KVP: Magnetic easy axis",
        "E_x": "KVP: Soc. total energy, x-direction [eV/unit cell]",
        "E_y": "KVP: Soc. total energy, y-direction [eV/unit cell]",
        "E_z": "KVP: Soc. total energy, z-direction [eV/unit cell]",
        "theta": "Easy axis, polar coordinates, theta [radians]",
        "phi": "Easy axis, polar coordinates, phi [radians]",
        "dE_zx":
        "KVP: Magnetic anisotropy energy between x and z axis [meV/unit cell]",
        "dE_zy":
        "KVP: Magnetic anisotropy energy between y and z axis [meV/unit cell]",
    },
    "pdos": {
        "pdos_nosoc":
        "Projected density of states w/o soc. !PDOS no soc!",
        "pdos_soc":
        "Projected density of states !PDOS!",
        "dos_at_ef_nosoc":
        "KVP: Density of states at the Fermi energy w/o soc."
        "!DOS at ef no soc.! [states/(eV * unit cell)]",
        "dos_at_ef_soc":
        "KVP: Density of states at the Fermi energy"
        " !DOS at ef! [states/(eV * unit cell)]",
    },
    "phonons": {
        "minhessianeig": "KVP: Minimum eigenvalue of Hessian [`eV/Ang^2`]",
        "dynamic_stability_level": "KVP: Dynamic stability level",
    },
    "plasmafrequency": {
        "plasmafreq_vv": "Plasma frequency tensor [Hartree]",
        "plasmafrequency_x": "KVP: 2D plasma frequency (x)"
        "[`eV/Ang^0.5`]",
        "plasmafrequency_y": "KVP: 2D plasma frequency (y)"
        "[`eV/Ang^0.5`]",
    },
    "polarizability": {
        "alphax_el": "KVP: Static electronic polarizability (x) [Ang]",
        "alphay_el": "KVP: Static electronic polarizability (y) [Ang]",
        "alphaz_el": "KVP: Static electronic polarizability (z) [Ang]",
    },
    "relax": {
        "etot": "Total energy [eV]",
        "edft": "DFT total enrgy [eV]",
        "spos": "Array: Scaled positions",
        "symbols": "Array: Chemical symbols",
        "a": "Cell parameter a [Ang]",
        "b": "Cell parameter b [Ang]",
        "c": "Cell parameter c [Ang]",
        "alpha": "Cell parameter alpha [deg]",
        "beta": "Cell parameter beta [deg]",
        "gamma": "Cell parameter gamma [deg]",
    },
    "stiffness": {
        "c_11": "KVP: Stiffness tensor, 11-component [`N/m^{dim-1}`]",
        "c_22": "KVP: Stiffness tensor, 22-component [`N/m^{dim-1}`]",
        "c_33": "KVP: Stiffness tensor, 33-component [`N/m^{dim-1}`]",
        "c_23": "KVP: Stiffness tensor, 23-component [`N/m^{dim-1}`]",
        "c_13": "KVP: Stiffness tensor, 13-component [`N/m^{dim-1}`]",
        "c_12": "KVP: Stiffness tensor, 12-component [`N/m^{dim-1}`]",
        "speed_of_sound_x": "KVP: Speed of sound (x) [m/s]",
        "speed_of_sound_y": "KVP: Speed of sound (y) [m/s]",
        "stiffness_tensor": "Stiffness tensor [`N/m^{dim-1}`]",
    },
    "structureinfo": {
        "magstate": "KVP: Magnetic state",
        "is_magnetic": "KVP: Material is magnetic !Magnetic!",
        "cell_area": "KVP: Area of unit-cell [`Ang^2`]",
        "has_inversion_symmetry": "KVP: Material has inversion symmetry",
        "stoichiometry": "KVP: Stoichiometry",
        "spacegroup": "KVP: Space group",
        "spgnum": "KVP: Space group number",
        "crystal_prototype": "KVP: Crystal prototype",
    },
    "database.material_fingerprint": {
        'asr_id': 'KVP: Material unique ID',
        'uid': 'KVP: Unique identifier'
    },
    "info.json": {
        'class': 'KVP: Material class',
        'doi': 'KVP: Monolayer reported DOI',
        'icsd_id': 'KVP: ICSD id of parent bulk structure',
        'cod_id': 'KVP: COD id of parent bulk structure'
    },
    "emasses": {
        'emass_vb_dir1':
        'KVP: Valence band effective mass, direction 1 [`m_e`]',
        'emass_vb_dir2':
        'KVP: Valence band effective mass, direction 2 [`m_e`]',
        'emass_vb_dir3':
        'KVP: Valence band effective mass, direction 3 [`m_e`]',
        'emass_cb_dir1':
        'KVP: Conduction band effective mass, direction 1 [`m_e`]',
        'emass_cb_dir2':
        'KVP: Conduction band effective mass, direction 2 [`m_e`]',
        'emass_cb_dir3':
        'KVP: Conduction band effective mass, direction 3 [`m_e`]',
    },
    "database.fromtree": {
        "folder": "KVP: Relative path to original folder",
    }
}

# Piezoelectrictensor key_descriptions
piezokd = {}
for i in range(1, 4):
    for j in range(1, 7):
        key = 'e_{}{}'.format(i, j)
        name = 'Piezoelectric tensor'
        description = f'Piezoelectric tensor {i}{j}-component' + '[`\\text{Ang}^{-1}`]'
        piezokd[key] = description

key_descriptions['piezoelectrictensor'] = piezokd

# Key descriptions like has_asr_gs_calculate
extras = {}
for recipe in get_recipes():
    key = 'has_' + recipe.name.replace('.', '_').replace('@', '_')
    extras[key] = f'{recipe.name} is calculated'

key_descriptions['extra'] = extras


@command()
@argument('database')
def main(database):
    """Analyze database and set metadata.

    This recipe loops through all rows in a database and figures out what keys
    are present in the database. It also figures our what kind of materials
    (1D, 2D, 3D) are in the database. Then it saves those values in the
    database metadata under {"keys": ["etot", "gap", ...]}
    """
    from ase.db import connect

    db = connect(database)

    print('Row #')
    keys = set()
    for ir, row in enumerate(db.select(include_data=False)):
        if ir % 100 == 0:
            print(ir)
        print(row.key_value_pairs.keys())
        keys.update(set(row.key_value_pairs.keys()))

    metadata = db.metadata
    metadata.update({'keys': sorted(list(keys))})
    db.metadata = metadata


if __name__ == '__main__':
    main.cli()



from asr.core import command, argument, option


@command('asr.database.merge')
@argument('databaseout', nargs=1)
@argument('databases', nargs=-1)
@option('--identifier', help='Identifier for matching database entries.')
def main(databases, databaseout, identifier='uid'):
    """Merge two ASE databases."""
    from ase.db import connect
    from pathlib import Path
    from click import progressbar

    print(f'Merging {databases} into {databaseout}')

    def item_show_func(item):
        if item is None:
            return item
        return str(item.formula)

    dest = Path(databaseout)
    assert not dest.is_file(), \
        f'The destination path {databaseout} already exists.'

    # We build up a temporary database file at this destination
    tmpdest = Path(databaseout + '.tmp.db')
    if tmpdest.is_file():
        tmpdest.unlink()

    # First merge rows common in both databases
    metadata = {'keys': set()}
    for database in databases:
        # Database for looking up existing materials
        tmpdestsearch = Path('_' + str(tmpdest))
        if tmpdestsearch.is_file():
            tmpdestsearch.unlink()

        if tmpdest.is_file():
            tmpdest.rename(tmpdestsearch)

        print(f'Connecting to {database} (may take a while)', flush=True)
        db = connect(database)
        dbsearch = connect(str(tmpdestsearch))
        dbmerged = connect(str(tmpdest))

        with dbmerged, progressbar(db.select(),
                                   length=len(db),
                                   label=f'Merging {database}',
                                   item_show_func=item_show_func) as selection:
            id_matches = []
            for row1 in selection:
                structid = row1.get(identifier)
                matching = list(dbsearch.select(f'{identifier}={structid}'))

                if len(matching) > 1:
                    raise RuntimeError('More than one structure '
                                       f'in {databaseout} '
                                       f'matching {identifier}={structid}')
                elif len(matching) == 0:
                    atoms = row1.toatoms()
                    kvp = row1.key_value_pairs
                    data = row1.data
                else:
                    row2 = matching[0]
                    id_matches.append(row2.id)
                    data = row2.data
                    kvp = row2.key_value_pairs

                    data.update(row1.data.copy())
                    kvp.update(row1.key_value_pairs.copy())

                    atoms = row1.toatoms()
                    atoms2 = row2.toatoms()
                    assert atoms == atoms2, 'Atoms not matching!'

                dbmerged.write(atoms,
                               key_value_pairs=kvp,
                               data=data)

            # Write the remaining rows from db2 that wasn't matched
            for row2 in dbsearch.select():
                if row2.id not in id_matches:
                    dbmerged.write(row2.toatoms(),
                                   key_value_pairs=row2.key_value_pairs,
                                   data=row2.data)

        # Update metadata
        metadata['keys'].update(set(db.metadata.get('keys', [])))

    metadata['keys'] = sorted(list(metadata['keys']))
    dbmerged.metadata = metadata

    # Remove lookup db
    tmpdestsearch.unlink()

    # Copy the file to the final destination
    tmpdest.rename(dest)


if __name__ == '__main__':
    main.cli()



from asr.core import command, option

tests = [{'cli': ['ase build -x diamond Si structure.json',
                  'asr run "setup.strains --kptdensity 2.0"',
                  'asr run "setup.params asr.relax:calculator '
                  '''{'mode':{'ecut':200},'kpts':(1,1,1),...}" strains*/''',
                  'asr run relax strains*/',
                  'asr run database.material_fingerprint strains*/',
                  'asr run stiffness']}]


def webpanel(row, key_descriptions):
    import numpy as np

    def matrixtable(M, digits=2, unit='', skiprow=0, skipcolumn=0):
        table = M.tolist()
        shape = M.shape

        for i in range(skiprow, shape[0]):
            for j in range(skipcolumn, shape[1]):
                value = table[i][j]
                table[i][j] = '{:.{}f}{}'.format(value, digits, unit)
        return table
    stiffnessdata = row.data['results-asr.stiffness.json']
    c_ij = stiffnessdata['stiffness_tensor']
    eigs = stiffnessdata['eigenvalues']

    c_ij = np.zeros((4, 4))
    c_ij[1:, 1:] = stiffnessdata['stiffness_tensor']
    rows = matrixtable(c_ij, unit='',
                       skiprow=1,
                       skipcolumn=1)
    rows[0] = ['C<sub>ij</sub> (N/m)', 'xx', 'yy', 'xy']
    rows[1][0] = 'xx'
    rows[2][0] = 'yy'
    rows[3][0] = 'xy'
    for ir, tmprow in enumerate(rows):
        for ic, item in enumerate(tmprow):
            if ir == 0 or ic == 0:
                rows[ir][ic] = '<b>' + rows[ir][ic] + '</b>'

    ctable = dict(
        type='table',
        rows=rows)

    eigrows = ([['<b>Stiffness tensor eigenvalues<b>', '']]
               + [[f'Eigenvalue {ie}', f'{eig:.2f} N/m']
                  for ie, eig in enumerate(sorted(eigs,
                                                  key=lambda x: x.real))])
    eigtable = dict(
        type='table',
        rows=eigrows)

    panel = {'title': 'Stiffness tensor',
             'columns': [[ctable], [eigtable]],
             'sort': 2}

    dynstab = ['low', 'high'][int(eigs.min() > 0)]
    high = 'Min. Stiffness eig. > 0'
    low = 'Min. Stiffness eig. < 0'
    row = ['Dynamical (stiffness)',
           '<a href="#" data-toggle="tooltip" data-html="true" '
           + 'title="LOW: {}&#13;HIGH: {}">{}</a>'.format(
               low, high, dynstab.upper())]

    summary = {'title': 'Summary',
               'columns': [[{'type': 'table',
                             'header': ['Stability', 'Category'],
                             'rows': [row],
                             }]],
               'sort': 3}

    return [panel, summary]


@command(module='asr.stiffness',
         webpanel=webpanel,
         tests=tests)
@option('--strain-percent', help='Magnitude of applied strain')
def main(strain_percent=1.0):
    from asr.setup.strains import (get_strained_folder_name,
                                   get_relevant_strains)
    from ase.io import read
    from ase.units import J
    from asr.core import read_json, chdir
    from asr.database.material_fingerprint import main as computemf
    import numpy as np

    atoms = read('structure.json')
    ij = get_relevant_strains(atoms.pbc)

    ij_to_voigt = [[0, 5, 4],
                   [5, 1, 3],
                   [4, 3, 2]]

    links = {}
    stiffness = np.zeros((6, 6), float)
    for i, j in ij:
        dstress = np.zeros((6,), float)
        for sign in [-1, 1]:
            folder = get_strained_folder_name(sign * strain_percent, i, j)
            structurefile = folder / 'structure.json'
            with chdir(folder):
                if not computemf.done:
                    computemf()
            mf = read_json(folder / ('results-asr.database.'
                                     'material_fingerprint.json'))
            links[str(folder)] = mf['uid']
            structure = read(str(structurefile))
            # The structure already has the stress if it was
            # calculated
            stress = structure.get_stress(voigt=True)
            dstress += stress * sign
        stiffness[:, ij_to_voigt[i][j]] = dstress / (strain_percent * 0.02)

    stiffness = np.array(stiffness, float)
    # We work with Mandel notation which is conventional and convenient
    stiffness[3:, :] *= 2**0.5
    stiffness[:, 3:] *= 2**0.5

    # Convert the stiffness tensor from [eV/Ang^3] -> [J/m^3]=[N/m^2]
    stiffness *= 10**30 / J

    # Now do some post processing
    data = {'__key_descriptions__': {}}
    kd = data['__key_descriptions__']
    nd = np.sum(atoms.pbc)
    if nd == 2:
        cell = atoms.get_cell()
        # We have to normalize with the supercell size
        z = cell[2, 2]
        stiffness = stiffness[[0, 1, 5], :][:, [0, 1, 5]] * z * 1e-10
        from ase.units import kg
        from ase.units import m as meter
        area = atoms.get_volume() / cell[2, 2]
        mass = sum(atoms.get_masses())
        area_density = (mass / kg) / (area / meter**2)
        # speed of sound in m/s
        speed_x = np.sqrt(stiffness[0, 0] / area_density)
        speed_y = np.sqrt(stiffness[1, 1] / area_density)
        data['speed_of_sound_x'] = speed_x
        data['speed_of_sound_y'] = speed_y
        data['c_11'] = stiffness[0, 0]
        data['c_22'] = stiffness[1, 1]
        data['c_33'] = stiffness[2, 2]
        data['c_23'] = stiffness[1, 2]
        data['c_13'] = stiffness[0, 2]
        data['c_12'] = stiffness[0, 1]
        kd['c_11'] = 'KVP: Stiffness tensor: 11-component [N/m]'
        kd['c_22'] = 'KVP: Stiffness tensor: 22-component [N/m]'
        kd['c_33'] = 'KVP: Stiffness tensor: 33-component [N/m]'
        kd['c_23'] = 'KVP: Stiffness tensor: 23-component [N/m]'
        kd['c_13'] = 'KVP: Stiffness tensor: 13-component [N/m]'
        kd['c_12'] = 'KVP: Stiffness tensor: 12-component [N/m]'
        kd['speed_of_sound_x'] = 'KVP: Speed of sound in x direction [m/s]'
        kd['speed_of_sound_y'] = 'KVP: Speed of sound in y direction [m/s]'
        kd['stiffness_tensor'] = 'Stiffness tensor [N/m]'
    elif nd == 1:
        area = atoms.get_volume() / cell[2, 2]
        stiffness = stiffness[5, 5] * area * 1e-20
        kd['stiffness_tensor'] = 'Stiffness tensor [N]'
    else:
        kd['stiffness_tensor'] = 'Stiffness tensor [N/m^2]'

    data['__links__'] = links
    data['stiffness_tensor'] = stiffness

    eigs = np.linalg.eigvals(stiffness)
    data['eigenvalues'] = eigs
    return data


if __name__ == '__main__':
    main.cli()



from asr.core import command


def todict(atoms):
    import numpy as np
    d = dict(atoms.arrays)
    d['cell'] = np.asarray(atoms.cell)
    d['pbc'] = atoms.pbc
    if atoms._celldisp.any():
        d['celldisp'] = atoms._celldisp
    # if atoms.constraints:
    #     d['constraints'] = atoms.constraints
    if atoms.info:
        d['info'] = atoms.info
    return d


@command(module='asr.database.material_fingerprint')
def main():
    import numpy as np
    from hashlib import md5
    import json
    from ase.io import read
    from collections import OrderedDict

    atoms = read('structure.json')
    dct = todict(atoms)

    for key, value in dct.items():
        if isinstance(value, np.ndarray):
            value = value.tolist()
        dct[key] = value

    # Make sure that that keys appear in order
    orddct = OrderedDict()
    keys = list(dct.keys())
    keys.sort()
    for key in keys:
        orddct[key] = dct[key]

    hash = md5(json.dumps(orddct).encode()).hexdigest()
    formula = atoms.symbols.formula
    results = {'asr_id': hash,
               'uid': f'{formula:abc}-' + hash[:12]}
    results['__key_descriptions__'] = {'asr_id': 'KVP: Material fingerprint',
                                       'uid': 'KVP: Unique identifier'}
    return results


if __name__ == '__main__':
    main.cli()



from .conftest import test_materials
from asr.relax import BrokenSymmetryError
from pathlib import Path
import pytest
import numpy as np


@pytest.mark.ci
@pytest.mark.parametrize("atoms", test_materials)
def test_relax(separate_folder, mockgpaw, atoms):
    from asr.relax import main as relax
    from ase.io import write

    write('unrelaxed.json', atoms)
    relax(calculator={
        "name": "gpaw",
        "kpts": {"density": 2, "gamma": True},
    })


@pytest.mark.ci
@pytest.mark.parametrize('name', ['Al', 'Cu', 'Ag', 'Au', 'Ni',
                                  'Pd', 'Pt', 'C'])
def test_relax_emt(separate_folder, name):
    from asr.relax import main as relax
    from ase.build import bulk

    unrelaxed = bulk(name)
    unrelaxed.write('unrelaxed.json')
    relax(calculator={'name': 'emt'})


@pytest.mark.ci
@pytest.mark.parametrize('name', ['Al', 'Cu', 'Ag', 'Au', 'Ni',
                                  'Pd', 'Pt', 'C'])
def test_relax_emt_fail_broken_symmetry(separate_folder, name,
                                        monkeypatch, capsys):
    """Test that a broken symmetry raises an error."""
    from asr.relax import main as relax
    from ase.build import bulk
    from ase.calculators.emt import EMT

    unrelaxed = bulk(name)

    def get_stress(*args, **kwargs):
        return np.random.rand(3, 3)

    monkeypatch.setattr(EMT, 'get_stress', get_stress)
    unrelaxed.write('unrelaxed.json')
    with pytest.raises(BrokenSymmetryError) as excinfo:
        relax(calculator={'name': 'emt'}, enforce_symmetry=False)

    assert 'The symmetry was broken during the relaxation!' in str(excinfo.value)


@pytest.mark.ci
def test_relax_find_higher_symmetry(separate_folder, monkeypatch, capsys):
    """Test that a structure is allowed to find a higher symmetry without failing."""
    from ase.build import bulk
    from asr.relax import main, SpgAtoms, myBFGS

    diamond = bulk('C')
    natoms = len(diamond)
    sposoriginal_ac = diamond.get_scaled_positions()
    spos_ac = diamond.get_scaled_positions()
    spos_ac[1][2] += 0.1
    diamond.set_scaled_positions(spos_ac)

    def get_stress(*args, **kwargs):
        return np.zeros((6,), float)

    def get_forces(*args, **kwargs):
        return 1 + np.zeros((natoms, 3), float)

    def irun(self, *args, **kwargs):
        yield False
        self.atoms.atoms.set_scaled_positions(sposoriginal_ac)
        yield False

    diamond.write('unrelaxed.json')

    monkeypatch.setattr(SpgAtoms, 'get_forces', get_forces)
    monkeypatch.setattr(SpgAtoms, 'get_stress', get_stress)
    monkeypatch.setattr(myBFGS, 'irun', irun)
    main(calculator={'name': 'emt'})

    captured = capsys.readouterr()
    assert "The spacegroup has changed during relaxation. " in captured.out


@pytest.mark.integration_test
@pytest.mark.integration_test_gpaw
def test_relax_si_gpaw(separate_folder):
    from asr.setup.materials import main as setupmaterial
    from asr.relax import main as relax
    setupmaterial.cli(["-s", "Si2"])
    Path("materials.json").rename("unrelaxed.json")
    relaxargs = (
        "{'mode':{'ecut':200,'dedecut':'estimate',...},"
        "'kpts':{'density':1,'gamma':True},...}"
    )
    results = relax.cli(["--calculator", relaxargs])
    assert abs(results["c"] - 3.978) < 0.001


@pytest.mark.integration_test
@pytest.mark.integration_test_gpaw
def test_relax_bn_gpaw(separate_folder):
    from asr.setup.materials import main as setupmaterial
    from asr.relax import main as relax
    from asr.core import read_json

    setupmaterial.cli(["-s", "BN,natoms=2"])
    Path("materials.json").rename("unrelaxed.json")
    relaxargs = (
        "{'mode':{'ecut':300,'dedecut':'estimate',...},"
        "'kpts':{'density':2,'gamma':True},...}"
    )
    relax.cli(["--calculator", relaxargs])

    results = read_json("results-asr.relax.json")
    assert results["c"] > 5



def get_kpts_size(atoms, density):
    """Try to get a reasonable Monkhorst-Pack grid which hits high symmetry points."""
    from gpaw.kpt_descriptor import kpts2sizeandoffsets as k2so
    size, offset = k2so(atoms=atoms, density=density)
    # size[2] = 1  # what do we do in general? XXX
    for i in range(len(size)):
        if size[i] % 6 != 0 and size[i] != 1:  # works for hexagonal cells XXX
            size[i] = 6 * (size[i] // 6 + 1)
    kpts = {'size': size, 'gamma': True}
    return kpts



import numpy as np
from asr.core import command, option, read_json


@command(module='asr.berry',
         requires=['gs.gpw'],
         dependencies=['asr.gs@calculate'],
         resources='120:10h')
@option('--gs', help='Ground state')
@option('--kpar', help='K-points along path')
@option('--kperp', help='K-points orthogonal to path')
def calculate(gs='gs.gpw', kpar=120, kperp=7):
    """Calculate ground state on specified k-point grid."""
    import os
    from ase.io import read
    from gpaw import GPAW
    from gpaw.berryphase import parallel_transport
    from gpaw.mpi import world

    atoms = read('structure.json')
    pbc = atoms.pbc.tolist()

    """Find the easy axis of magnetic materials"""
    theta = 0.0
    phi = 0.0
    if os.path.isfile('results-asr.magnetic_anisotropy.json'):
        a = read_json('results-asr.magnetic_anisotropy.json')
        dE_zy = a['dE_zy']
        dE_zx = a['dE_zx']
        if dE_zy > 0 or dE_zx > 0:
            theta = np.pi / 2
            if dE_zy > dE_zx:
                phi = np.pi / 2

    ND = np.sum(pbc)

    results = {}

    if ND == 2:
        calc = GPAW(gs,
                    kpts=(kperp, kpar, 1),
                    fixdensity=True,
                    symmetry='off',
                    txt='gs_berry.txt')
        calc.get_potential_energy()
        calc.write('gs_berry.gpw', mode='all')
        phi_km, s_km = parallel_transport('gs_berry.gpw',
                                          direction=0,
                                          theta=theta,
                                          phi=phi)
        results['phi0_km'] = phi_km
        results['s0_km'] = s_km

        if world.rank == 0:
            os.system('rm gs_berry.gpw')

        return results

    elif ND == 3:
        """kx = 0"""
        calc = GPAW(gs,
                    kpts=(1, kperp, kpar),
                    fixdensity=True,
                    symmetry='off',
                    txt='gs_berry.txt')
        calc.get_potential_energy()
        calc.write('gs_berry.gpw', mode='all')
        phi_km, s_km = parallel_transport('gs_berry.gpw',
                                          direction=1,
                                          theta=theta,
                                          phi=phi)
        results['phi1_km'] = phi_km
        results['s1_km'] = s_km

        """ky = 0"""
        calc.set(kpts=(kpar, 1, kperp))
        calc.get_potential_energy()
        calc.write('gs_berry.gpw', mode='all')
        phi_km, s_km = parallel_transport('gs_berry.gpw',
                                          direction=2,
                                          theta=theta,
                                          phi=phi)
        results['phi2_km'] = phi_km
        results['s2_km'] = s_km

        """kz = 0"""
        calc.set(kpts=(kperp, kpar, 1))
        calc.get_potential_energy()
        calc.write('gs_berry.gpw', mode='all')
        phi_km, s_km = parallel_transport('gs_berry.gpw',
                                          direction=0,
                                          theta=theta,
                                          phi=phi)
        results['phi0_km'] = phi_km
        results['s0_km'] = s_km

        r"""kz = \pi"""
        from ase.dft.kpoints import monkhorst_pack
        kpts = monkhorst_pack((kperp, kpar, 1)) + [0, 0, 0.5]
        calc.set(kpts=kpts)
        calc.get_potential_energy()
        calc.write('gs_berry.gpw', mode='all')
        phi_km, s_km = parallel_transport('gs_berry.gpw',
                                          direction=0,
                                          theta=theta,
                                          phi=phi)
        results['phi0_pi_km'] = phi_km
        results['s0_pi_km'] = s_km

        if world.rank == 0:
            os.system('rm gs_berry.gpw')

        return results

    else:
        return


def plot_phases(name='0'):
    import pylab as plt

    results = read_json('results-asr.berry@calculate.json')
    phit_km = results['phi%s_km' % name]
    St_km = results['s%s_km' % name]
    Nk = len(St_km)

    phi_km = np.zeros((len(phit_km) + 1, len(phit_km[0])), float)
    phi_km[1:] = phit_km
    phi_km[0] = phit_km[-1]
    S_km = np.zeros((len(phit_km) + 1, len(phit_km[0])), float)
    S_km[1:] = St_km
    S_km[0] = St_km[-1]
    S_km /= 2

    Nm = len(phi_km[0])
    phi_km = np.tile(phi_km, (1, 2))
    phi_km[:, Nm:] += 2 * np.pi
    S_km = np.tile(S_km, (1, 2))

    plt.figure()
    plt.scatter(np.tile(np.arange(len(phi_km)), len(phi_km.T)),
                phi_km.T.reshape(-1),
                cmap=plt.get_cmap('viridis'),
                c=S_km.T.reshape(-1),
                s=5,
                marker='o')

    cbar = plt.colorbar()
    cbar.set_label(r'$\langle S_z\rangle/\hbar$', size=20)

    plt.ylabel(r'$\gamma_x$', size=24)
    plt.xticks([0, Nk / 2, Nk],
               [r'$-\mathrm{M}$', r'$\Gamma$', r'$\mathrm{M}$'], size=20)
    plt.yticks([0, np.pi, 2 * np.pi], [r'$0$', r'$\pi$', r'$2\pi$'], size=20)
    plt.axis([0, Nk, 0, 2 * np.pi])
    plt.tight_layout()
    plt.show()


def webpanel(row, key_descriptions):
    if row.Topology == 'Not checked':
        return []

    row = ['Band topology', row.Topology]
    summary = {'title': 'Summary',
               'columns': [[{'type': 'table',
                             'header': ['Electronic properties', ''],
                             'rows': [row]}]]}

    basicelec = {'title': 'Basic electronic properties (PBE)',
                 'columns': [[{'type': 'table',
                               'header': ['Property', ''],
                               'rows': [row]}]],
                 'sort': 15}

    return [summary, basicelec]


@command(module='asr.berry',
         requires=['results-asr.berry@calculate.json'],
         dependencies=['asr.berry@calculate'],
         webpanel=webpanel)
def main():
    from pathlib import Path
    from ase.parallel import paropen

    data = {}
    if Path('topology.dat').is_file():
        f = paropen('topology.dat', 'r')
        top = f.readline()
        f.close()
        data['Topology'] = top
        data['__key_descriptions__'] = \
            {'Topology': 'KVP: Band topology (Topology)'}
    else:
        f = paropen('topology.dat', 'w')
        print('Not checked!', file=f)
        f.close()
        data['Topology'] = 'Not checked'

    return data


if __name__ == '__main__':
    main.cli()



import pytest

from click.testing import CliRunner
from asr.core.cli import cli


@pytest.mark.ci
def test_asr():
    """Test the main CLI."""
    runner = CliRunner()
    result = runner.invoke(cli)
    assert result.exit_code == 0
    assert 'Usage: cli [OPTIONS] COMMAND [ARGS]...' in result.output
    help_result = runner.invoke(cli, ['--help'])
    assert help_result.exit_code == 0
    assert '--help  Show this message and exit.' in help_result.output
    help_result = runner.invoke(cli, ['-h'])
    assert help_result.exit_code == 0
    assert '--help  Show this message and exit.' in help_result.output


@pytest.mark.ci
def test_asr_run(separate_folder):
    import pathlib
    runner = CliRunner()
    result = runner.invoke(cli, ['run', '-h'])
    assert result.exit_code == 0
    assert 'Usage: cli run [OPTIONS] COMMAND [FOLDERS]' in result.output

    help_result = runner.invoke(cli, ['--help'])
    assert help_result.exit_code == 0
    assert '--help  Show this message and exit.' in help_result.output

    help_result = runner.invoke(cli, ['-h'])
    assert help_result.exit_code == 0
    assert '--help  Show this message and exit.' in help_result.output

    result = runner.invoke(cli, ['run', '--dry-run', 'setup.params'])
    assert result.exit_code == 0
    assert 'Would run asr.setup.params@main in 1 folders.' in result.output

    pathlib.Path("folder1").mkdir()
    pathlib.Path("folder2").mkdir()

    result = runner.invoke(cli, ['run',
                                 'setup.params asr.relax:d3 True',
                                 'folder1', 'folder2'])
    assert result.exit_code == 0
    assert pathlib.Path("folder1", "params.json").is_file()
    assert pathlib.Path("folder2", "params.json").is_file()

    pathlib.Path('str1.json').write_text("")
    result = runner.invoke(cli, ['run', '--shell', 'mv str1.json str2.json'])
    assert pathlib.Path("str2.json").is_file()

    pathlib.Path("folder3").mkdir()
    pathlib.Path("folder4").mkdir()

    result = runner.invoke(cli, ['run', '--jobs', '2',
                                 'setup.params asr.relax:d3 True',
                                 'folder3', 'folder4'])
    assert result.exit_code == 0
    assert pathlib.Path("folder3", "params.json").is_file(), result
    assert pathlib.Path("folder4", "params.json").is_file(), result


@pytest.mark.ci
def test_asr_list():
    runner = CliRunner()
    result = runner.invoke(cli, ['list'])
    assert result.exit_code == 0
    assert 'Name' in result.output
    assert 'Description' in result.output


@pytest.mark.ci
def test_asr_results():
    runner = CliRunner()
    result = runner.invoke(cli, ['results', '-h'])
    assert result.exit_code == 0
    assert 'Usage: cli results [OPTIONS] NAME' in result.output


@pytest.mark.ci
def test_asr_find_help():
    runner = CliRunner()
    result = runner.invoke(cli, ['find', '-h'])
    assert result.exit_code == 0
    assert 'Usage: cli find [OPTIONS] RECIPE [HASH]...' in result.output


@pytest.mark.ci
@pytest.mark.parametrize(
    "recipe,hashish,output",
    [('asr.recipename', '9e2e1e68..32241753', 'results-asr.recipename.json\n'),
     ('asr.recipename', '^9e2e1e68 32241753', 'results-asr.recipename.json\n'),
     ('asr.recipename', 'c8980f6f3^..c8980f6f3', 'results-asr.recipename.json\n'),
     ('asr.recipename', '^c8980f6f3^ c8980f6f3', 'results-asr.recipename.json\n'),
     ('asr.recipename', 'c8980f6f3..c8980f6f3^', ''),
     ('asr.recipename', '^c8980f6f3 c8980f6f3^', ''),
     ('asr.recipename', 'c8980f6f3..32241753', ''),
     ('asr.recipename', '^c8980f6f3 32241753', '')])
def test_asr_find(recipe, hashish, output):
    from asr.core import write_json
    # TODO: Mock git call
    data = {
        '__versions__': {
            'asr':
            'version-c8980f6f32492437136b3b88b6d2598a8b653a25'
        }
    }
    recipe = "asr.recipename"
    filename = f'results-{recipe}.json'
    write_json(filename, data)

    runner = CliRunner()
    result = runner.invoke(
        cli,
        ['find', recipe] + hashish.split())

    assert result.exit_code == 0
    assert result.output == output


@pytest.mark.ci
def test_asr_find_no_versions(separate_folder):
    from asr.core import write_json
    data = {'dummydata': ['somecontent']}
    recipe = "asr.recipename"
    filename = f'results-{recipe}.json'
    write_json(filename, data)

    runner = CliRunner()
    result = runner.invoke(
        cli,
        ['find', recipe, 'c8980f6f3..32241753'])

    assert result.exit_code == 0
    assert result.output == ''



import pytest
from pytest import approx
from .conftest import test_materials, get_webcontent


@pytest.mark.ci
@pytest.mark.parametrize("atoms", test_materials)
@pytest.mark.parametrize("gap", [0, 1])
@pytest.mark.parametrize("fermi_level", [0.5, 1.5])
def test_gs(separate_folder, mockgpaw, mocker, atoms, gap, fermi_level):
    from asr.gs import calculate, main
    from ase.io import write
    from ase.units import Ha
    import gpaw
    import gpaw.occupations
    mocker.patch.object(gpaw.GPAW, "_get_band_gap")
    mocker.patch.object(gpaw.GPAW, "_get_fermi_level")
    mocker.patch("gpaw.occupations.occupation_numbers")
    gpaw.GPAW._get_fermi_level.return_value = fermi_level
    gpaw.GPAW._get_band_gap.return_value = gap
    gpaw.occupations.occupation_numbers.return_value = [0,
                                                        fermi_level / Ha,
                                                        0,
                                                        0]

    write('structure.json', atoms)
    calculate(
        calculator={
            "name": "gpaw",
            "kpts": {"density": 2, "gamma": True},
        },
    )

    results = main()
    assert results.get("gaps_nosoc").get("efermi") == approx(fermi_level)
    assert results.get("efermi") == approx(fermi_level)
    if gap >= fermi_level:
        assert results.get("gap") == approx(gap)
    else:
        assert results.get("gap") == approx(0)

    content = get_webcontent('database.db')
    resultgap = results.get("gap")
    assert f"<td>Bandgap</td><td>{resultgap:0.2f}eV</td>" in content, content
    assert f"<td>Fermilevel</td><td>{fermi_level:0.3f}eV</td>" in \
        content, content


@pytest.mark.ci
def test_gs_asr_cli_results_figures(separate_folder, mockgpaw):
    from pathlib import Path
    from asr.gs import main
    from asr.core.material import (get_material_from_folder,
                                   get_webpanels_from_material,
                                   make_panel_figures)
    atoms = test_materials[0]
    atoms.write('structure.json')

    main()
    material = get_material_from_folder()
    panel = get_webpanels_from_material(material, main)
    make_panel_figures(material, panel)
    assert Path('bz-with-gaps.png').is_file()



import pytest
import os
import numpy as np
import contextlib
from pathlib import Path

from ase import Atoms
from ase.build import bulk


def get_webcontent(name='database.db'):
    from asr.database.fromtree import main as fromtree
    fromtree()

    from asr.database import app as appmodule
    from pathlib import Path
    from asr.database.app import app, initialize_project, projects

    tmpdir = Path("tmp/")
    tmpdir.mkdir()
    appmodule.tmpdir = tmpdir
    initialize_project(name)

    app.testing = True
    with app.test_client() as c:
        content = c.get(f"/database.db/").data.decode()
        project = projects["database.db"]
        db = project["database"]
        uid_key = project["uid_key"]
        row = db.get(id=1)
        uid = row.get(uid_key)
        url = f"/database.db/row/{uid}"
        content = c.get(url).data.decode()
        content = (
            content
            .replace("\n", "")
            .replace(" ", "")
        )
    return content


@pytest.fixture()
def mockgpaw(monkeypatch):
    import sys
    monkeypatch.syspath_prepend(Path(__file__).parent.resolve() / "mocks")
    for module in list(sys.modules):
        if "gpaw" in module:
            sys.modules.pop(module)

    yield sys.path

    for module in list(sys.modules):
        if "gpaw" in module:
            sys.modules.pop(module)

# @pytest.fixture()
# def mockgpaw(mocker):
#     from sys import modules
#     from .mocks import gpaw
#     mocker.patch.dict(
#         modules,
#         {
#             "gpaw": gpaw,
#         })


# Make some 1D, 2D and 3D test materials
Si = bulk("Si")
Ag = bulk("Ag")
Ag2 = bulk("Ag").repeat((2, 1, 1))
Fe = bulk("Fe")
Fe.set_initial_magnetic_moments([1])
abn = 2.51
BN = Atoms(
    "BN",
    scaled_positions=[[0, 0, 0.5], [1 / 3, 2 / 3, 0.5]],
    cell=[
        [abn, 0.0, 0.0],
        [-0.5 * abn, np.sqrt(3) / 2 * abn, 0],
        [0.0, 0.0, 15.0],
    ],
    pbc=[True, True, False],
)
Agchain = Atoms(
    "Ag",
    scaled_positions=[[0.5, 0.5, 0]],
    cell=[
        [15.0, 0.0, 0.0],
        [0.0, 15.0, 0.0],
        [0.0, 0.0, 2],
    ],
    pbc=[False, False, True],
)
test_materials = [Si, BN, Agchain]


@contextlib.contextmanager
def create_new_working_directory(path='workdir', unique=False):
    """Change working directory and returns to previous on exit."""
    i = 0
    if unique:
        while Path(f'{path}-{i}').is_dir():
            i += 1
        path = f'{path}-{i}'

    Path(path).mkdir()
    prev_cwd = Path.cwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(prev_cwd)


@pytest.fixture()
def separate_folder(tmpdir):
    """Create temp folder and change directory to that folder.

    A context manager that creates a temporary folder and changes
    the current working directory to it for isolated filesystem tests.
    """
    cwd = os.getcwd()
    os.chdir(str(tmpdir))

    try:
        yield str(tmpdir)
    finally:
        os.chdir(cwd)


def pytest_configure(config):
    # register an additional marker
    config.addinivalue_line(
        "markers",
        """integration_test: Marks an integration test""",
    )
    config.addinivalue_line(
        "markers",
        """integration_test_gpaw: Marks an integration
        test specifically using gpaw""",
    )
    config.addinivalue_line(
        "markers",
        """ci: Mark a test for running in continuous integration""",
    )


def freeelectroneigenvalues(atoms, gap=0):
    def get_eigenvalues(self, kpt, spin=0):
        from ase.calculators.calculator import kpts2ndarray
        from ase.units import Bohr, Ha
        import numpy as np

        if hasattr(self, "tmpeigenvalues"):
            return self.tmpeigenvalues[kpt]
        nelectrons = self.get_number_of_electrons()
        kpts = kpts2ndarray(self.parameters.kpts, atoms)
        nbands = self.get_number_of_bands()

        icell = atoms.get_reciprocal_cell() * 2 * np.pi * Bohr

        # Simple parabolic band
        n = 3
        offsets = np.indices((n, n, n)).T.reshape((n ** 3, 1, 3)) - n // 2
        eps_kn = 0.5 * (np.dot(kpts + offsets, icell) ** 2).sum(2).T
        eps_kn.sort()

        eps_kn = np.concatenate(
            (-eps_kn[:, ::-1][:, -nelectrons:],
             eps_kn + gap / Ha),
            axis=1,
        )

        self.tmpeigenvalues = eps_kn[:, : nbands] * Ha
        return self.tmpeigenvalues[kpt]
    return get_eigenvalues



from pathlib import Path

from asr.utils.kpts import get_kpts_size


def nonselfc(txt=None, kptdensity=20.0, emptybands=20):
    """Non self-consistent calculation based on the density in gs.gpw."""
    from gpaw import GPAW
    calc = GPAW('gs.gpw', txt=None)
    spinpol = calc.get_spin_polarized()

    kpts = get_kpts_size(atoms=calc.atoms, density=kptdensity)
    convbands = int(emptybands / 2)
    calc.set(nbands=-emptybands,
             txt=txt,
             fixdensity=True,
             kpts=kpts,
             convergence={'bands': -convbands})

    if spinpol:
        calc.set(symmetry='off')  # due to soc

    calc.get_potential_energy()

    return calc


def get_filenames(gpw, txt, selfc=False, **kwargs):
    """Get file names as specified by input."""
    parstr = get_parstr(selfc=selfc, **kwargs)

    if isinstance(gpw, str):
        if gpw == 'default':
            gpw = f'refinedgs_{parstr}.gpw'
        assert gpw[-4:] == '.gpw'
    else:
        assert gpw is None

    if isinstance(txt, str):
        if txt == 'default':
            txt = f'refinedgs_{parstr}.txt'
    else:
        assert txt is None

    return gpw, txt


def get_parstr(selfc=False, **kwargs):
    """Get parameter string, specifying how the ground state is refined."""
    parstr = 'selfc«%s»' % str(selfc)

    for kw in ['kptdensity', 'emptybands']:
        parstr += '_%s«%s»' % (kw, str(kwargs[kw]))

    return parstr


def refinegs(selfc=False, gpw=None, txt=None, **kwargs):
    """Refine the ground state calculation.

    Parameters
    ----------
    selfc : bool
        Perform new self-consistency cycle to refine also the density
    gpw : str
        Write the refined ground state as a .gpw file.
        If 'default' is specified, use f'refinedgs_{parstr}.gpw' as file name.
        If another string is specified, use that as file name.
    txt : str
        Write the GPAW output to a .txt file.
        If 'default' is specified, use f'refinedgs_{parstr}.txt' as file name.
        If another string is specified, use that as file name.

    Returns
    -------
    calc : obj
        GPAW calculator object
    gpw : str
        filename of written GPAW calculator object
    """
    from gpaw import GPAW
    gpw, txt = get_filenames(gpw, txt, selfc=selfc, **kwargs)
    if gpw and Path(gpw).is_file():
        calc = GPAW(gpw, txt=None)
    else:
        if selfc:
            raise NotImplementedError('Someone should implement refinement '
                                      + 'with self-consistency')
        else:
            calc = nonselfc(txt=txt, **kwargs)

        if gpw:
            calc.write(gpw)

    return calc, gpw



from asr.core import command, option


def get_kpts_size(atoms, density):
    """Try to get a reasonable monkhorst size which hits high symmetry points."""
    from gpaw.kpt_descriptor import kpts2sizeandoffsets as k2so
    size, offset = k2so(atoms=atoms, density=density)
    size[2] = 1
    for i in range(2):
        if size[i] % 6 != 0:
            size[i] = 6 * (size[i] // 6 + 1)
    kpts = {'size': size, 'gamma': True}
    return kpts


@command('asr.plasmafrequency',
         creates=['es_plasma.gpw'])
@option('--kptdensity', help='k-point density')
def calculate(kptdensity=20):
    """Calculate excited states for polarizability calculation."""
    from gpaw import GPAW

    calc_old = GPAW('gs.gpw', txt=None)
    kpts = get_kpts_size(atoms=calc_old.atoms, density=kptdensity)
    nval = calc_old.wfs.nvalence
    calc = GPAW('gs.gpw', fixdensity=True, kpts=kpts,
                nbands=2 * nval, txt='gsplasma.txt')
    calc.get_potential_energy()
    calc.write('es_plasma.gpw', 'all')


def webpanel(row, key_descriptions):
    from asr.database.browser import table

    if row.get('gap', 1) > 0.01:
        return []

    plasmatable = table(row, 'Property', [
        'plasmafrequency_x', 'plasmafrequency_y'], key_descriptions)

    panel = {'title': 'Optical polarizability (RPA)',
             'columns': [[], [plasmatable]]}
    return [panel]


@command('asr.plasmafrequency',
         requires=['es_plasma.gpw'],
         webpanel=webpanel,
         dependencies=['asr.plasmafrequency@calculate'])
@option('--tetra', is_flag=True,
        help='Use tetrahedron integration')
def main(tetra=True):
    """Calculate polarizability."""
    from gpaw.response.df import DielectricFunction
    from ase.io import read
    import numpy as np
    from ase.units import Hartree, Bohr

    atoms = read('structure.json')
    nd = sum(atoms.pbc)
    if not nd == 2:
        raise AssertionError('Plasmafrequency recipe only implemented for 2D')

    if tetra:
        kwargs = {'truncation': '2D',
                  'eta': 0.05,
                  'domega0': 0.2,
                  'integrationmode': 'tetrahedron integration',
                  'ecut': 1,
                  'pbc': [True, True, False]}
    else:
        kwargs = {'truncation': '2D',
                  'eta': 0.05,
                  'domega0': 0.2,
                  'ecut': 1}

    df = DielectricFunction('es_plasma.gpw', **kwargs)
    df.get_polarizability(q_c=[0, 0, 0], direction='x',
                          pbc=[True, True, False],
                          filename=None)
    plasmafreq_vv = df.chi0.plasmafreq_vv.real
    data = {'plasmafreq_vv': plasmafreq_vv,
            '__key_descriptions__': {'plasmafreq_vv':
                                     'Plasma frequency tensor [Hartree]'}}

    if nd == 2:
        wp2_v = np.linalg.eigvalsh(plasmafreq_vv[:2, :2])
        L = atoms.cell[2, 2] / Bohr
        plasmafreq_v = (np.sqrt(wp2_v * L / 2) * Hartree * Bohr**0.5)
        data['plasmafrequency_x'] = plasmafreq_v[0].real
        data['plasmafrequency_y'] = plasmafreq_v[1].real

        data['__key_descriptions__']['plasmafrequency_x'] = \
            'KVP: 2D plasma frequency, x-direction [eV/Ang^0.5]'
        data['__key_descriptions__']['plasmafrequency_y'] = \
            'KVP: 2D plasma frequency, y-direction [eV/Ang^0.5]'

    return data


if __name__ == '__main__':
    main.cli()



from asr.core import command, option
import sys
from pathlib import Path
from typing import List, Dict, Tuple, Any
import traceback

import matplotlib.pyplot as plt
from ase.db.row import AtomsRow
from ase.db.core import float_to_time_string, now

assert sys.version_info >= (3, 4)

plotlyjs = (
    '<script src="https://cdn.plot.ly/plotly-latest.min.js">' + '</script>')
external_libraries = [plotlyjs]

unique_key = 'uid'

params = {'legend.fontsize': 'large',
          'axes.labelsize': 'large',
          'axes.titlesize': 'large',
          'xtick.labelsize': 'large',
          'ytick.labelsize': 'large',
          'savefig.dpi': 200}
plt.rcParams.update(**params)


def create_table(row,  # AtomsRow
                 header,  # List[str]
                 keys,  # List[str]
                 key_descriptions,  # Dict[str, Tuple[str, str, str]]
                 digits=3  # int
                 ):  # -> Dict[str, Any]
    """Create table-dict from row."""
    table = []
    for key in keys:
        if key == 'age':
            age = float_to_time_string(now() - row.ctime, True)
            table.append(('Age', age))
            continue
        value = row.get(key)
        if value is not None:
            if isinstance(value, float):
                value = '{:.{}f}'.format(value, digits)
            elif not isinstance(value, str):
                value = str(value)
            desc, unit = key_descriptions.get(key, ['', key, ''])[1:]
            if unit:
                value += ' ' + unit
            table.append((desc, value))
    return {'type': 'table',
            'header': header,
            'rows': table}


def miscellaneous_section(row, key_descriptions, exclude):
    """Make help function for adding a "miscellaneous" section.

    Create table with all keys except those in exclude.
    """
    misckeys = (set(key_descriptions)
                | set(row.key_value_pairs)) - set(exclude)
    misc = create_table(row, ['Items', ''], sorted(misckeys), key_descriptions)
    return ('Miscellaneous', [[misc]])


def val2str(row, key: str, digits=2) -> str:
    value = row.get(key)
    if value is not None:
        if isinstance(value, float):
            value = '{:.{}f}'.format(value, digits)
        elif not isinstance(value, str):
            value = str(value)
    else:
        value = ''
    return value


def fig(filename: str, link: str = None) -> 'Dict[str, Any]':
    """Shortcut for figure dict."""
    dct = {'type': 'figure', 'filename': filename}
    if link:
        dct['link'] = link
    return dct


def table(row, title, keys, kd={}, digits=2):
    return create_table(row, [title, 'Value'], keys, kd, digits)


def merge_panels(page):
    """Merge panels which have the same title.

    Also merge tables with same first entry in header.
    """
    # Update panels
    for title, panels in page.items():
        panels = sorted(panels, key=lambda x: x['sort'])

        panel = {'title': title,
                 'columns': [[], []],
                 'plot_descriptions': [],
                 'sort': panels[0]['sort']}
        known_tables = {}
        for tmppanel in panels:
            for column in tmppanel['columns']:
                for ii, item in enumerate(column):
                    if isinstance(item, dict):
                        if item['type'] == 'table':
                            if 'header' not in item:
                                continue
                            header = item['header'][0]
                            if header in known_tables:
                                known_tables[header]['rows']. \
                                    extend(item['rows'])
                                column[ii] = None
                            else:
                                known_tables[header] = item

            columns = tmppanel['columns']
            if len(columns) == 1:
                columns.append([])

            columns[0] = [item for item in columns[0] if item]
            columns[1] = [item for item in columns[1] if item]
            panel['columns'][0].extend(columns[0])
            panel['columns'][1].extend(columns[1])
            panel['plot_descriptions'].extend(tmppanel['plot_descriptions'])
        page[title] = panel


def layout(row: AtomsRow,
           key_descriptions: Dict[str, Tuple[str, str, str]],
           prefix: Path) -> List[Tuple[str, List[List[Dict[str, Any]]]]]:
    """Page layout."""
    from asr.core import get_recipes
    page = {}
    exclude = set()

    # Locate all webpanels
    recipes = get_recipes()
    for recipe in recipes:
        if not recipe.webpanel:
            continue
        # We assume that there should be a results file in
        if f'results-{recipe.name}.json' not in row.data:
            continue
        try:
            panels = recipe.webpanel(row, key_descriptions)
        except Exception:
            traceback.print_exc()
            panels = []
        for thispanel in panels:
            assert 'title' in thispanel, f'No title in {recipe.name} webpanel'
            panel = {'columns': [[], []],
                     'plot_descriptions': [],
                     'sort': 99}
            panel.update(thispanel)
            paneltitle = panel['title']
            if paneltitle in page:
                page[paneltitle].append(panel)
            else:
                page[paneltitle] = [panel]

    merge_panels(page)
    page = [panel for _, panel in page.items()]
    # Sort sections if they have a sort key
    page = [x for x in sorted(page, key=lambda x: x.get('sort', 99))]

    misc_title, misc_columns = miscellaneous_section(row, key_descriptions,
                                                     exclude)
    misc_panel = {'title': misc_title,
                  'columns': misc_columns}
    page.append(misc_panel)

    # Get descriptions of figures that are created by all webpanels
    plot_descriptions = []
    for panel in page:
        plot_descriptions.extend(panel.get('plot_descriptions', []))

    # List of functions and the figures they create:
    missing = set()  # missing figures
    for desc in plot_descriptions:
        function = desc['function']
        filenames = desc['filenames']
        paths = [Path(prefix + filename) for filename in filenames]
        for path in paths:
            if not path.is_file():
                # Create figure(s) only once:
                try:
                    function(row, *(str(path) for path in paths))
                except Exception:
                    traceback.print_exc()
                plt.close('all')
                for path in paths:
                    if not path.is_file():
                        path.write_text('')  # mark as missing
                break
        for path in paths:
            if path.stat().st_size == 0:
                missing.add(path)

    # We convert the page into ASE format
    asepage = []
    for panel in page:
        asepage.append((panel['title'], panel['columns']))

    def ok(block):
        if block is None:
            return False
        if block['type'] == 'table':
            return block['rows']
        if block['type'] != 'figure':
            return True
        if Path(prefix + block['filename']) in missing:
            return False
        return True

    # Remove missing figures from layout:
    final_page = []
    for title, columns in asepage:
        columns = [[block for block in column if ok(block)]
                   for column in columns]
        if any(columns):
            final_page.append((title, columns))
    return final_page


@command('asr.database.browser')
@option('--database')
@option('--only-figures', is_flag=True,
        help='Dont show browser, just save figures')
def main(database='database.db', only_figures=False):
    """Open results in web browser."""
    import subprocess
    from pathlib import Path

    custom = Path(__file__)

    cmd = f'python3 -m ase db {database} -w -M {custom}'
    if only_figures:
        cmd += ' -l'
    print(cmd)
    try:
        subprocess.check_output(cmd.split())
    except subprocess.CalledProcessError as e:
        print(e.output)
        exit(1)


if __name__ == '__main__':
    main.cli()



from asr.core import command, read_json


def get_spin_axis():
    anis = read_json('results-asr.magnetic_anisotropy.json')
    return anis['theta'], anis['phi']


def get_spin_index():
    anis = read_json('results-asr.magnetic_anisotropy.json')
    axis = anis['spin_axis']
    if axis == 'z':
        index = 2
    elif axis == 'y':
        index = 1
    else:
        index = 0
    return index


def spin_axis(theta, phi):
    import numpy as np
    if theta == 0:
        return 'z'
    elif np.allclose(phi, np.pi / 2):
        return 'y'
    else:
        return 'x'


def webpanel(row, key_descriptions):
    from asr.database.browser import table
    if row.get('magstate', 'NM') == 'NM':
        return []

    magtable = table(row, 'Property',
                     ['magstate', 'magmom',
                      'dE_zx', 'dE_zy'], kd=key_descriptions)
    panel = {'title': 'Basic magnetic properties (PBE)',
             'columns': [[magtable], []],
             'sort': 11}
    return [panel]


params = '''asr.gs@calculate:calculator +{'mode':'lcao','kpts':(2,2,2)}'''
tests = [{'cli': ['ase build -x hcp Co structure.json',
                  f'asr run "setup.params {params}"',
                  'asr run asr.magnetic_anisotropy',
                  'asr run database.fromtree',
                  'asr run "database.browser --only-figures"']}]


@command('asr.magnetic_anisotropy',
         tests=tests,
         requires=['gs.gpw', 'results-asr.structureinfo.json'],
         webpanel=webpanel,
         dependencies=['asr.gs@calculate', 'asr.structureinfo'])
def main():
    """Calculate the magnetic anisotropy.

    Uses the magnetic anisotropy to calculate the preferred spin orientation
    for magnetic (FM/AFM) systems.

    Returns
    -------
        theta: Polar angle in radians
        phi: Azimuthal angle in radians
    """
    import numpy as np
    from asr.core import file_barrier, read_json
    from gpaw.mpi import world, serial_comm
    from gpaw.spinorbit import get_anisotropy
    from gpaw import GPAW
    from gpaw.utilities.ibz2bz import ibz2bz
    from pathlib import Path

    structureinfo = read_json('results-asr.structureinfo.json')
    magstate = structureinfo['magstate']
    # Figure out if material is magnetic
    results = {'__key_descriptions__':
               {'spin_axis': 'KVP: Suggested spin direction for SOC',
                'E_x': 'KVP: SOC total energy difference in x-direction',
                'E_y': 'KVP: SOC total energy difference in y-direction',
                'E_z': 'KVP: SOC total energy difference in z-direction',
                'theta': 'Spin direction, theta, polar coordinates [radians]',
                'phi': 'Spin direction, phi, polar coordinates [radians]',
                'dE_zx': ('KVP: Magnetic anisotropy energy '
                          '(zx-component) [meV/formula unit]'),
                'dE_zy': ('KVP: Magnetic anisotropy energy '
                          '(zy-component) [meV/formula unit]')}}

    if magstate == 'NM':
        results['E_x'] = 0
        results['E_y'] = 0
        results['E_z'] = 0
        results['dE_zx'] = 0
        results['dE_zy'] = 0
        results['theta'] = 0
        results['phi'] = 0
        results['spin_axis'] = 'z'
        return results

    with file_barrier(['gs_nosym.gpw']):
        ibz2bz('gs.gpw', 'gs_nosym.gpw')
    width = 0.001
    nbands = None
    calc = GPAW('gs_nosym.gpw', communicator=serial_comm, txt=None)
    E_x = get_anisotropy(calc, theta=np.pi / 2, nbands=nbands,
                         width=width)
    calc = GPAW('gs_nosym.gpw', communicator=serial_comm, txt=None)
    E_z = get_anisotropy(calc, theta=0.0, nbands=nbands, width=width)
    calc = GPAW('gs_nosym.gpw', communicator=serial_comm, txt=None)
    E_y = get_anisotropy(calc, theta=np.pi / 2, phi=np.pi / 2,
                         nbands=nbands, width=width)

    dE_zx = E_z - E_x
    dE_zy = E_z - E_y

    DE = max(dE_zx, dE_zy)
    theta = 0
    phi = 0
    if DE > 0:
        theta = np.pi / 2
        if dE_zy > dE_zx:
            phi = np.pi / 2

    axis = spin_axis(theta, phi)

    results.update({'spin_axis': axis,
                    'theta': theta,
                    'phi': phi,
                    'E_x': E_x * 1e3,
                    'E_y': E_y * 1e3,
                    'E_z': E_z * 1e3,
                    'dE_zx': dE_zx * 1e3,
                    'dE_zy': dE_zy * 1e3})
    world.barrier()
    if world.rank == 0:
        Path('gs_nosym.gpw').unlink()
    return results


if __name__ == '__main__':
    main.cli()



import numpy as np
from asr.core import command, option
from ase.dft.bandgap import bandgap
from click import Choice


def get_kpts_size(atoms, kptdensity):
    """Try to get a reasonable monkhorst size which hits high symmetry points."""
    from gpaw.kpt_descriptor import kpts2sizeandoffsets as k2so
    size, offset = k2so(atoms=atoms, density=kptdensity)
    size[2] = 1
    for i in range(2):
        if size[i] % 6 != 0:
            size[i] = 6 * (size[i] // 6 + 1)
    kpts = {'size': size, 'gamma': True}
    return kpts


@command(creates=['bse_polx.csv', 'bse_eigx.dat',
                  'bse_poly.csv', 'bse_eigy.dat',
                  'bse_polz.csv', 'bse_eigz.dat'],
         requires=['gs.gpw'],
         resources='480:20h')
@option('--gs', help='Ground state on which BSE is based')
@option('--kptdensity', help='K-point density')
@option('--ecut', help='Plane wave cutoff')
@option('--nv_s', help='Valence bands included')
@option('--nc_s', help='Conduction bands included')
@option('--mode', help='Irreducible response',
        type=Choice(['RPA', 'BSE', 'TDHF']))
@option('--bandfactor', type=int,
        help='Number of unoccupied bands = (#occ. bands) * bandfactor)')
def calculate(gs='gs.gpw', kptdensity=6.0, ecut=50.0, mode='BSE', bandfactor=6,
              nv_s=-2.3, nc_s=2.3):
    """Calculate BSE polarizability."""
    import os
    from ase.io import read
    from gpaw import GPAW
    from gpaw.mpi import world
    from gpaw.response.bse import BSE
    from gpaw.occupations import FermiDirac
    from pathlib import Path
    import numpy as np
    from asr.core import file_barrier

    atoms = read('structure.json')
    pbc = atoms.pbc.tolist()
    ND = np.sum(pbc)
    if ND == 3:
        eta = 0.1
        kpts = {'density': kptdensity, 'gamma': True, 'even': True}
        truncation = None
    elif ND == 2:
        eta = 0.05
        kpts = get_kpts_size(atoms=atoms, kptdensity=20)
        truncation = '2D'

    else:
        raise NotImplementedError(
            'asr for BSE not implemented for 0D and 1D structures')

    calc_gs = GPAW(gs, txt=None)
    spin = calc_gs.get_spin_polarized()
    nval = calc_gs.wfs.nvalence
    nocc = int(nval / 2)
    nbands = bandfactor * nocc
    Nk = len(calc_gs.get_ibz_k_points())
    gap, v, c = bandgap(calc_gs, direct=True, output=None)

    if isinstance(nv_s, float):
        ev = calc_gs.get_eigenvalues(kpt=v[1], spin=v[0])[v[2]]
        nv_sk = np.zeros((spin + 1, Nk), int)
        for s in range(spin + 1):
            for k in range(Nk):
                e_n = calc_gs.get_eigenvalues(kpt=k, spin=s)
                e_n -= ev
                x = e_n[np.where(e_n < 0)]
                x = x[np.where(x > nv_s)]
                nv_sk[s, k] = len(x)
        nv_s = np.max(nv_sk, axis=1)
    if isinstance(nc_s, float):
        ec = calc_gs.get_eigenvalues(kpt=c[1], spin=c[0])[c[2]]
        nc_sk = np.zeros((spin + 1, Nk), int)
        for s in range(spin + 1):
            for k in range(Nk):
                e_n = calc_gs.get_eigenvalues(kpt=k, spin=s)
                e_n -= ec
                x = e_n[np.where(e_n > 0)]
                x = x[np.where(x < nc_s)]
                nc_sk[s, k] = len(x)
        nc_s = np.max(nc_sk, axis=1)

    nv_s = [np.max(nv_s), np.max(nv_s)]
    nc_s = [np.max(nc_s), np.max(nc_s)]
    print('nv_s, nc_s', nv_s, nc_s)
    valence_bands = []
    conduction_bands = []
    for s in range(spin + 1):
        gap, v, c = bandgap(calc_gs, direct=True, spin=s, output=None)
        valence_bands.append(range(c[2] - nv_s[s], c[2]))
        conduction_bands.append(range(c[2], c[2] + nc_s[s]))

    print(valence_bands)
    print(conduction_bands)

    if not Path('gs_bse.gpw').is_file():
        calc = GPAW(
            gs,
            txt='gs_bse.txt',
            fixdensity=True,
            nbands=int(nbands * 1.5),
            convergence={'bands': nbands},
            occupations=FermiDirac(width=1e-4),
            kpts=kpts)
        calc.get_potential_energy()
        with file_barrier(['gs_bse.gpw']):
            calc.write('gs_bse.gpw', mode='all')

    # if spin:
    #     f0 = calc.get_occupation_numbers(spin=0)
    #     f1 = calc.get_occupation_numbers(spin=1)
    #     n0 = np.where(f0 < 1.0e-6)[0][0]
    #     n1 = np.where(f1 < 1.0e-6)[0][0]
    #     valence_bands = [range(n0 - nv, n0), range(n1 - nv, n1)]
    #     conduction_bands = [range(n0, n0 + nc), range(n1, n1 + nc)]
    # else:
    #     valence_bands = range(nocc - nv, nocc)
    #     conduction_bands = range(nocc, nocc + nc)

    world.barrier()

    bse = BSE('gs_bse.gpw',
              spinors=True,
              ecut=ecut,
              valence_bands=valence_bands,
              conduction_bands=conduction_bands,
              nbands=nbands,
              mode=mode,
              truncation=truncation,
              txt='bse.txt')

    w_w = np.linspace(-2.0, 8.0, 10001)

    w_w, alphax_w = bse.get_polarizability(eta=eta,
                                           filename='bse_polx.csv',
                                           direction=0,
                                           write_eig='bse_eigx.dat',
                                           pbc=pbc,
                                           w_w=w_w)

    w_w, alphay_w = bse.get_polarizability(eta=eta,
                                           filename='bse_poly.csv',
                                           direction=1,
                                           write_eig='bse_eigy.dat',
                                           pbc=pbc,
                                           w_w=w_w)

    w_w, alphaz_w = bse.get_polarizability(eta=eta,
                                           filename='bse_polz.csv',
                                           direction=2,
                                           write_eig='bse_eigz.dat',
                                           pbc=pbc,
                                           w_w=w_w)
    if world.rank == 0:
        os.system('rm gs_bse.gpw')
        os.system('rm gs_nosym.gpw')


def absorption(row, filename, direction='x'):
    import numpy as np
    import matplotlib.pyplot as plt
    from ase.units import alpha, Ha, Bohr

    atoms = row.toatoms()
    pbc = atoms.pbc.tolist()
    dim = np.sum(pbc)

    magstate = row.magstate

    gap_dir = row.gap_dir
    gap_dir_nosoc = row.gap_dir_nosoc

    for method in ['_gw', '_hse', '_gllbsc', '']:
        gapkey = f'gap_dir{method}'
        if gapkey in row:
            gap_dir_x = row.get(gapkey)
            delta_bse = gap_dir_x - gap_dir
            delta_rpa = gap_dir_x - gap_dir_nosoc
            break

    qp_gap = gap_dir + delta_bse

    if magstate != 'NM':
        qp_gap = gap_dir_nosoc + delta_rpa
        delta_bse = delta_rpa

    ax = plt.figure().add_subplot(111)

    data = row.data['results-asr.bse.json'][f'bse_alpha{direction}_w']
    wbse_w = data[:, 0] + delta_bse
    absbse_w = 4 * np.pi * data[:, 2]
    if dim == 2:
        absbse_w *= wbse_w * alpha / Ha / Bohr * 100
    ax.plot(wbse_w, absbse_w, '-', c='0.0', label='BSE')
    xmax = wbse_w[-1]

    # TODO: Sometimes RPA pol doesn't exist, what to do?
    data = row.data.get('results-asr.polarizability.json')
    if data:
        wrpa_w = data['frequencies'] + delta_rpa
        absrpa_w = 4 * np.pi * data[f'alpha{direction}_w'].imag
        if dim == 2:
            absrpa_w *= wrpa_w * alpha / Ha / Bohr * 100
        ax.plot(wrpa_w, absrpa_w, '-', c='C0', label='RPA')
        ymax = max(np.concatenate([absbse_w[wbse_w < xmax],
                                   absrpa_w[wrpa_w < xmax]])) * 1.05
    else:
        ymax = max(absbse_w[wbse_w < xmax]) * 1.05
    ax.plot([qp_gap, qp_gap], [0, ymax], '--', c='0.5',
            label='Direct QP gap')

    ax.set_xlim(0.0, xmax)
    ax.set_ylim(0.0, ymax)
    ax.set_title(f'{direction}-polarization')
    ax.set_xlabel('energy [eV]')
    if dim == 2:
        ax.set_ylabel('Absorbance [%]')
    else:
        ax.set_ylabel(r'$\varepsilon(\omega)$')
    ax.legend()
    plt.tight_layout()
    plt.savefig(filename)

    return ax


def webpanel(row, key_descriptions):
    from functools import partial
    from asr.database.browser import fig, table

    E_B = table(row, 'Property', ['E_B'], key_descriptions)

    atoms = row.toatoms()
    pbc = atoms.pbc.tolist()
    dim = np.sum(pbc)

    if dim == 2:
        funcx = partial(absorption, direction='x')
        funcz = partial(absorption, direction='z')

        panel = {'title': 'Optical absorption (BSE and RPA)',
                 'columns': [[fig('absx.png'), E_B],
                             [fig('absz.png')]],
                 'plot_descriptions': [{'function': funcx,
                                        'filenames': ['absx.png']},
                                       {'function': funcz,
                                        'filenames': ['absz.png']}]}
    else:
        funcx = partial(absorption, direction='x')
        funcy = partial(absorption, direction='y')
        funcz = partial(absorption, direction='z')

        panel = {'title': 'Optical absorption (BSE and RPA)',
                 'columns': [[fig('absx.png'), fig('absz.png')],
                             [fig('absy.png'), E_B]],
                 'plot_descriptions': [{'function': funcx,
                                        'filenames': ['absx.png']},
                                       {'function': funcy,
                                        'filenames': ['absy.png']},
                                       {'function': funcz,
                                        'filenames': ['absz.png']}]}
    return [panel]


@command(module='asr.bse',
         requires=['bse_polx.csv', 'results-asr.gs.json',
                   'results-asr.structureinfo.json'],
         dependencies=['asr.bse@calculate', 'asr.gs', 'asr.structureinfo'],
         webpanel=webpanel)
def main():
    alphax_w = np.loadtxt('bse_polx.csv', delimiter=',')
    data = {'bse_alphax_w': alphax_w.astype(np.float32)}

    from pathlib import Path
    if Path('bse_poly.csv').is_file():
        alphay_w = np.loadtxt('bse_poly.csv', delimiter=',')
        data['bse_alphay_w'] = alphay_w.astype(np.float32)
    if Path('bse_polz.csv').is_file():
        alphaz_w = np.loadtxt('bse_polz.csv', delimiter=',')
        data['bse_alphaz_w'] = alphaz_w.astype(np.float32)
    from asr.core import read_json

    if Path('bse_eigx.dat').is_file():
        E = np.loadtxt('bse_eigx.dat')[0, 1]

        info = read_json('results-asr.structureinfo.json')
        magstate = info['magstate']

        gsresults = read_json('results-asr.gs.json')
        if magstate == 'NM':
            E_B = gsresults['gap_dir'] - E
        else:
            E_B = gsresults['gap_dir_nosoc'] - E

        data['E_B'] = E_B
        data['__key_descriptions__'] = \
            {'E_B': 'KVP: BSE binding energy (Exc. bind. energy) [eV]'}

    return data


if __name__ == '__main__':
    main.cli()



import pytest

from .conftest import test_materials, freeelectroneigenvalues, get_webcontent


@pytest.mark.ci
@pytest.mark.parametrize("atoms", test_materials)
def test_gw(separate_folder, atoms, mockgpaw, mocker):
    import numpy as np
    import gpaw
    from gpaw.response.g0w0 import G0W0

    get_eigenvalues = freeelectroneigenvalues(atoms, gap=1)
    mocker.patch.object(gpaw.GPAW, "get_eigenvalues", new=get_eigenvalues)
    mocker.patch.object(gpaw.GPAW, "get_fermi_level")
    gpaw.GPAW.get_fermi_level.return_value = 0.5

    from asr.gw import main
    atoms.write("structure.json")
    ndim = sum(atoms.pbc)

    def calculate(self):
        self.calc.get_eigenvalues(kpt=0)
        eps = self.calc.tmpeigenvalues[np.newaxis, :, 0:6]

        return {"qp": eps,
                "Z": np.zeros_like(eps) + 1,
                "eps": eps}

    mocker.patch.object(G0W0, "calculate", calculate)
    if ndim > 1:
        main()
        get_webcontent('database.db')
    else:
        with pytest.raises(NotImplementedError):
            main()



from asr.core import command, option


def get_relevant_strains(pbc):
    import numpy as np
    if np.sum(pbc) == 3:
        ij = ((0, 0), (1, 1), (2, 2), (1, 2), (0, 2), (0, 1))
    elif np.sum(pbc) == 2:
        ij = ((0, 0), (1, 1), (0, 1))
    elif np.sum(pbc) == 1:
        ij = ((2, 2), )
    return ij


def get_strained_folder_name(strain_percent, i, j, clamped=False):
    from pathlib import Path
    import numpy as np
    if strain_percent == 0:
        return Path('.')
    itov_i = ['x', 'y', 'z']
    name = itov_i[i] + itov_i[j]
    sign = ['', '+', '-'][int(np.sign(strain_percent))]
    strain_percent = abs(float(strain_percent))

    if clamped:
        return Path(f'strains-{sign}{strain_percent}%-{name}-clamped/')

    return Path(f'strains-{sign}{strain_percent}%-{name}/')


def setup_strains(strain_percent=1, kptdensity=6.0, copyparams=True, clamp_atoms=False):
    from ase.io import read
    from ase.parallel import world
    from pathlib import Path
    import numpy as np
    from asr.setup.params import main as setup_params
    from asr.core import chdir, read_json, write_json
    from ase.calculators.calculator import kpts2sizeandoffsets

    atoms = read('structure.json')
    ij = get_relevant_strains(atoms.pbc)
    cell_cv = atoms.get_cell()
    size, _ = kpts2sizeandoffsets(density=kptdensity, atoms=atoms)

    nk1, nk2, nk3 = size
    for i, j in ij:
        for sign in [-1, 1]:
            signed_strain = sign * strain_percent
            strain_vv = np.eye(3)
            strain_vv[i, j] += signed_strain / 100.0
            strain_vv = (strain_vv + strain_vv.T) / 2
            strained_cell_cv = np.dot(cell_cv, strain_vv)
            atoms.set_cell(strained_cell_cv, scale_atoms=True)
            folder = get_strained_folder_name(signed_strain, i, j, clamped=clamp_atoms)
            if world.rank == 0:
                folder.mkdir()
            if clamp_atoms:
                filename = str(folder / 'structure.json')
            else:
                filename = str(folder / 'unrelaxed.json')
            atoms.write(filename)
            if copyparams:
                paramsfile = Path('params.json')
                if paramsfile.is_file():
                    write_json(folder / 'params.json', read_json(paramsfile))

            with chdir(folder):
                params = {
                    'asr.relax': {
                        'calculator': {
                            'kpts': {
                                'size': size,
                                'gamma': True,
                            },
                            None: None,
                        },
                        'fixcell': True,
                        'allow_symmetry_breaking': True,
                        'fmax': 0.008,
                    }
                }
                setup_params(params=params)


@command('asr.setup.strains')
@option('--strain-percent', help='Strain percentage')
@option('--kptdensity', help='Setup up relax and gs calc with fixed density')
@option('--copyparams/--dontcopyparams',
        help='Copy params.json from current folder into strained folders',
        is_flag=True)
def clamped(strain_percent=1, kptdensity=6.0, copyparams=True):
    results = setup_strains(strain_percent=strain_percent, kptdensity=kptdensity,
                            copyparams=copyparams, clamp_atoms=True)
    return results


@command('asr.setup.strains')
@option('--strain-percent', help='Strain percentage')
@option('--kptdensity', help='Setup up relax and gs calc with fixed density')
@option('--copyparams/--dontcopyparams',
        help='Copy params.json from current folder into strained folders',
        is_flag=True)
def main(strain_percent=1, kptdensity=6.0, copyparams=True):
    results = setup_strains(strain_percent=strain_percent, kptdensity=kptdensity,
                            copyparams=copyparams, clamp_atoms=False)
    return results


if __name__ == '__main__':
    main.cli()
